# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining BraketDeviceWrapper Class

"""
import json
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, List, Tuple

from braket.device_schema import ExecutionDay

from qbraid._qprogram import QPROGRAM_LIBS
from qbraid.providers.device import QuantumDevice
from qbraid.providers.enums import DeviceStatus, DeviceType

from .job import BraketQuantumTask

if TYPE_CHECKING:
    import braket.aws

    import qbraid


def _future_utc_datetime(hours: int, minutes: int, seconds: int) -> str:
    """Return a UTC datetime that is hours, minutes, and seconds from now
    as an ISO string without fractional seconds."""
    current_utc = datetime.utcnow()
    future_time = current_utc + timedelta(hours=hours, minutes=minutes, seconds=seconds)
    return future_time.strftime("%Y-%m-%dT%H:%M:%SZ")


class BraketDevice(QuantumDevice):
    """Wrapper class for Amazon Braket ``Device`` objects."""

    def __init__(self, aws_device: "braket.aws.AwsDevice"):
        """Create a BraketDevice."""

        super().__init__(aws_device)
        self._vendor = "AWS"
        self._run_package = "braket"

    def _populate_metadata(self, device: "braket.aws.AwsDevice") -> None:
        """Populate device metadata using BraketSchemaBase"""
        # pylint: disable=attribute-defined-outside-init
        metadata = device.aws_session.get_device(device.arn)
        capabilities = json.loads(metadata.get("deviceCapabilities"))

        self._id = metadata.get("deviceArn")
        self._name = metadata.get("deviceName")
        self._provider = metadata.get("providerName")
        self._device_type = DeviceType(metadata.get("deviceType"))

        try:
            self._num_qubits = capabilities["paradigm"]["qubitCount"]
        except KeyError:
            self._num_qubits = None

    def status(self) -> "qbraid.providers.DeviceStatus":
        """Return the status of this Device.

        Returns:
            The status of this Device
        """
        if self._device.is_available:
            return DeviceStatus.ONLINE

        if self._device.status == "RETIRED":
            return DeviceStatus.RETIRED

        return DeviceStatus.OFFLINE

    def availability_window(self) -> Tuple[bool, str, str]:
        """Provides device availability status. Indicates current availability,
        time remaining (hours, minutes, seconds) until next availability or
        unavailability, and future UTC datetime of next change in availability status.

        Returns:
            Tuple[bool, str, str]: Current device availability, hr/min/sec until availability
                                   switch, future UTC datetime of availability switch
        """

        is_available_result = False
        available_time = None

        device = self._device

        if device.status != "ONLINE":
            return is_available_result, "", ""

        day = 0

        current_datetime_utc = datetime.utcnow()
        for execution_window in device.properties.service.executionWindows:
            weekday = current_datetime_utc.weekday()
            current_time_utc = current_datetime_utc.time().replace(microsecond=0)

            if (
                execution_window.windowEndHour < execution_window.windowStartHour
                and current_time_utc < execution_window.windowEndHour
            ):
                weekday = (weekday - 1) % 7

            matched_day = execution_window.executionDay == ExecutionDay.EVERYDAY
            matched_day = matched_day or (
                execution_window.executionDay == ExecutionDay.WEEKDAYS and weekday < 5
            )
            matched_day = matched_day or (
                execution_window.executionDay == ExecutionDay.WEEKENDS and weekday > 4
            )
            ordered_days = (
                ExecutionDay.MONDAY,
                ExecutionDay.TUESDAY,
                ExecutionDay.WEDNESDAY,
                ExecutionDay.THURSDAY,
                ExecutionDay.FRIDAY,
                ExecutionDay.SATURDAY,
                ExecutionDay.SUNDAY,
            )
            matched_day = matched_day or (
                execution_window.executionDay in ordered_days
                and ordered_days.index(execution_window.executionDay) == weekday
            )

            matched_time = (
                execution_window.windowStartHour < execution_window.windowEndHour
                and execution_window.windowStartHour
                <= current_time_utc
                <= execution_window.windowEndHour
            ) or (
                execution_window.windowEndHour < execution_window.windowStartHour
                and (
                    current_time_utc >= execution_window.windowStartHour
                    or current_time_utc <= execution_window.windowEndHour
                )
            )

            if execution_window.executionDay in ordered_days:
                day = (ordered_days.index(execution_window.executionDay) - weekday) % 6

            start_time = execution_window.windowStartHour
            end_time = current_time_utc

            if day == 0:
                day_factor = day
            elif start_time.hour < end_time.hour:
                day_factor = day - 1
            elif start_time.hour == end_time.hour:
                if start_time.minute < end_time.minute:
                    day_factor = day - 1
                elif start_time.minute == end_time.minute:
                    if start_time.second <= end_time.second:
                        day_factor = day - 1
            else:
                day_factor = day

            td = timedelta(
                hours=start_time.hour,
                minutes=start_time.minute,
                seconds=start_time.second,
            ) - timedelta(hours=end_time.hour, minutes=end_time.minute, seconds=end_time.second)
            days_seconds = 86400 * day_factor
            available_time_secs = td.seconds + days_seconds

            if available_time is None or available_time_secs < available_time:
                available_time = available_time_secs

            is_available_result = is_available_result or (matched_day and matched_time)

        if available_time is None:
            return is_available_result, "", ""

        hours = available_time // 3600
        minutes = (available_time // 60) % 60
        seconds = available_time - hours * 3600 - minutes * 60
        available_time_hms = f"{hours:02}:{minutes:02}:{seconds:02}"
        utc_datetime_str = _future_utc_datetime(hours, minutes, seconds)
        return is_available_result, available_time_hms, utc_datetime_str

    def queue_depth(self) -> int:
        """Return the number of jobs in the queue for the device"""
        aws_device = self._device
        queue_depth_info = aws_device.queue_depth()
        queued_tasks = queue_depth_info.quantum_tasks
        num_queued_normal = queued_tasks["Normal"]
        num_queued_priority = queued_tasks["Priority"]
        return int(num_queued_normal) + int(num_queued_priority)

    def _transpile(self, run_input):
        """Transpile a circuit for the device."""
        if self._device_type.name == "SIMULATOR":
            # pylint: disable=import-outside-toplevel
            from qbraid.programs.braket import BraketCircuit

            program = BraketCircuit(run_input)
            program.remove_idle_qubits()
            run_input = program.program

        return run_input

    def _compile(self, run_input):
        """Compile a circuit for the device."""
        if self.provider.lower() == "ionq" and "pytket" in QPROGRAM_LIBS:
            # pylint: disable=import-outside-toplevel
            from qbraid.compiler.braket.ionq import braket_ionq_compile

            run_input = braket_ionq_compile(run_input)
        return run_input

    def run(
        self, run_input, *args, auto_compile: bool = False, **kwargs
    ) -> "qbraid.device.aws.BraketQuantumTaskWrapper":
        """Run a quantum task specification on this quantum device. Task must represent a
        quantum circuit, annealing problems not supported.

        Args:
            run_input: Specification of a task to run on device.

        Keyword Args:
            auto_compile (bool): Whether to compile the circuit for the device before running.
            shots (int): The number of times to run the task on the device.

        Returns:
            The job like object for the run.

        """
        qbraid_circuit = self.process_run_input(run_input, auto_compile=auto_compile)
        run_input = qbraid_circuit._program
        aws_quantum_task = self._device.run(run_input, *args, **kwargs)
        metadata = aws_quantum_task.metadata()
        shots = metadata.get("shots", 0)
        tags = metadata.get("tags", {})
        vendor_job_id = metadata["quantumTaskArn"]
        job_id = self._init_job(vendor_job_id, [qbraid_circuit], shots, tags)
        return BraketQuantumTask(
            job_id, vendor_job_id=vendor_job_id, device=self, vendor_job_obj=aws_quantum_task
        )

    def run_batch(
        self, run_input, *args, auto_compile: bool = False, **kwargs
    ) -> List["qbraid.device.aws.BraketQuantumTaskWrapper"]:
        """Run batch of quantum tasks on this quantum device.

        Args:
            run_input: A circuit object list to run on the wrapped device.

        Keyword Args:
            auto_compile (bool): Whether to compile the circuits for the device before running.
            shots (int): The number of times to run the task on the device. Default is 1024.

        Returns:
            List of BraketQuantumTask objects for the run.

        """
        device = self._device
        qbraid_circuit_batch = []
        run_input_batch = []
        for circuit in run_input:
            qbraid_circuit = self.process_run_input(circuit, auto_compile=auto_compile)
            run_input = qbraid_circuit._program
            run_input_batch.append(run_input)
            qbraid_circuit_batch.append(qbraid_circuit)
        aws_quantum_task_batch = device.run_batch(run_input_batch, *args, **kwargs)
        aws_quantum_tasks = aws_quantum_task_batch.tasks
        aws_quantum_task_wrapper_list = []
        for index, aws_quantum_task in enumerate(aws_quantum_tasks):
            qbraid_circuit = qbraid_circuit_batch[index]
            metadata = aws_quantum_task.metadata()
            shots = metadata.get("shots", 0)
            tags = metadata.get("tags", {})
            vendor_job_id = metadata["quantumTaskArn"]
            job_id = self._init_job(vendor_job_id, [qbraid_circuit], shots, tags)
            aws_quantum_task_wrapper_list.append(
                BraketQuantumTask(
                    job_id,
                    vendor_job_id=vendor_job_id,
                    device=self,
                    vendor_job_obj=aws_quantum_task,
                )
            )
        return aws_quantum_task_wrapper_list



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining CirqCircuit Class

"""

from typing import List

import cirq
import numpy as np

from qbraid.programs.abc_program import QuantumProgram


class CirqCircuit(QuantumProgram):
    """Wrapper class for ``cirq.Circuit`` objects."""

    def __init__(self, program: "cirq.Circuit"):
        super().__init__(program)

    @property
    def program(self) -> cirq.Circuit:
        return self._program

    @program.setter
    def program(self, value: cirq.Circuit) -> None:
        if not isinstance(value, cirq.Circuit):
            raise ValueError("Program must be an instance of cirq.Circuit")
        self._program = value

    @property
    def qubits(self) -> List[cirq.Qid]:
        """Return the qubits acted upon by the operations in this circuit"""
        return list(self.program.all_qubits())

    @property
    def num_clbits(self) -> int:
        """Return the number of classical bits in the circuit."""
        return 0

    @property
    def depth(self) -> int:
        """Return the circuit depth (i.e., length of critical path)."""
        return len(cirq.Circuit(self.program.all_operations()))

    def _unitary(self) -> np.ndarray:
        """Calculate unitary of circuit."""
        return self.program.unitary()

    @staticmethod
    def is_measurement_gate(op: cirq.Operation) -> bool:
        """Returns whether Cirq gate/operation is MeasurementGate."""
        return isinstance(op, cirq.ops.MeasurementGate) or isinstance(
            getattr(op, "gate", None), cirq.ops.MeasurementGate
        )

    @staticmethod
    def _key_from_qubit(qubit: cirq.Qid) -> str:
        if isinstance(qubit, cirq.LineQubit):
            key = str(qubit)
        elif isinstance(qubit, cirq.GridQubit):
            key = str(qubit.row)
        elif isinstance(qubit, cirq.NamedQubit):
            # Only correct if numbered sequentially
            key = str(qubit.name)
        else:
            raise ValueError(
                "Expected qubit of type 'GridQubit' 'LineQubit' or 'NamedQubit'"
                f"but instead got {type(qubit)}"
            )
        return key

    @staticmethod
    def _int_from_qubit(qubit: cirq.Qid) -> int:
        """Extracts an integer value from a given qubit.

        This function interprets the qubit based on its type and extracts an
        integer value accordingly. For a LineQubit, it returns its index.
        For a GridQubit, it returns the row number. For a NamedQubit, it attempts
        to parse and return an integer value from its name. This parsing is only
        successful if the name of the NamedQubit contains an integer. If the name
        does not contain an integer, such as in the case of NamedQubit("q"), or if
        the qubit is not of the supported types (LineQubit, GridQubit, or NamedQubit),
        a ValueError is raised.

        Args:
            qubit (cirq.Qid): The qubit from which to extract an integer.

        Returns:
            int: The extracted integer value from the qubit.

        Raises:
            ValueError: If the qubit is a NamedQubit whose name does not contain
                        an integer, or if the qubit is not of a supported
                        type (LineQubit, GridQubit, or NamedQubit).
        """
        if isinstance(qubit, cirq.LineQubit):
            return int(qubit)
        if isinstance(qubit, cirq.GridQubit):
            return qubit.row
        if isinstance(qubit, cirq.NamedQubit):
            try:
                # NamedQubit("q") will fail here
                return int(qubit._comparison_key().split(":")[0][7:])
            except ValueError as err:
                raise ValueError(
                    f"The provided NamedQubit '{qubit.name}' cannot be converted to an integer."
                ) from err
        else:
            raise ValueError(
                "Expected qubit of type 'GridQubit' 'LineQubit' or 'NamedQubit'"
                f"but instead got {type(qubit)}"
            )

    @staticmethod
    def _make_qubits(qubits: List[cirq.Qid], targets: List[int]) -> List[cirq.Qid]:
        if len(set(type(qubit) for qubit in qubits)) > 1:
            # If mixed types, default to LineQubits
            qubit_type = cirq.LineQubit
        else:
            qubit_type = type(qubits[0])

        if qubit_type == cirq.LineQubit:
            return [cirq.LineQubit(i) for i in targets]
        if qubit_type == cirq.GridQubit:
            return [cirq.GridQubit(0, i) for i in targets]
        if qubit_type == cirq.NamedQubit:
            return [cirq.NamedQubit(str(i)) for i in targets]
        raise ValueError(
            "Expected qubits of type 'GridQubit', 'LineQubit', or "
            f"'NamedQubit' but instead got {qubit_type}"
        )

    def _convert_to_line_qubits(self) -> None:
        """Converts a Cirq circuit constructed using NamedQubits to
        a Cirq circuit constructed using LineQubits."""
        qubits = self.qubits
        qubits.sort()
        try:
            qubit_map = {
                qubit: cirq.LineQubit(self._int_from_qubit(qubit)) for _, qubit in enumerate(qubits)
            }
        except ValueError:
            # If _int_from_qubit fails, fallback to default contiguous qubit mapping
            qubit_map = {qubit: cirq.LineQubit(i) for i, qubit in enumerate(qubits)}

        self._program = self.program.transform_qubits(lambda q: qubit_map[q])

    def populate_idle_qubits(self) -> None:
        """Checks whether the circuit uses contiguous qubits/indices,
        and if not, adds identity gates to vacant registers as needed."""
        circuit = self.program.copy()
        cirq_qubits = list(circuit.all_qubits())
        if any(isinstance(qubit, cirq.NamedQubit) for qubit in cirq_qubits):
            self._convert_to_line_qubits()
            self.populate_idle_qubits()
        else:
            nqubits = 0
            max_qubit = 0
            occupied_qubits = []
            for qubit in circuit.all_qubits():
                index = self._int_from_qubit(qubit)
                occupied_qubits.append(index)
                if index > max_qubit:
                    max_qubit = index
                nqubits += 1
            qubit_count = max_qubit + 1
            if qubit_count > nqubits:
                all_qubits = list(range(0, qubit_count))
                vacant_qubits = list(set(all_qubits) - set(occupied_qubits))
                cirq_qubits = self._make_qubits(cirq_qubits, vacant_qubits)
                for qubit in cirq_qubits:
                    circuit.append(cirq.I(qubit))
            self._program = circuit

    def remove_idle_qubits(self) -> None:
        """If circuit does not use contiguous qubits/indices,
        reduces dimension accordingly."""
        qubits = self.qubits
        qubits.sort()
        qubit_map = dict(zip(qubits, self._make_qubits(qubits, range(self.num_qubits))))
        self._program = self.program.transform_qubits(lambda q: qubit_map[q])

    def reverse_qubit_order(self) -> None:
        """Rerverse qubit ordering of circuit."""
        qubits = self.qubits
        qubits.sort()
        qubits_rev = qubits.copy()
        qubits_rev.reverse()
        qubit_map = dict(zip(qubits, qubits_rev))
        self._program = self.program.transform_qubits(lambda q: qubit_map[q])



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for BraketProvider class

"""
import os
import random
import string

import pytest
from braket.circuits import Circuit

from qbraid import device_wrapper, job_wrapper
from qbraid.providers.aws import BraketDevice, BraketProvider

# Skip tests if AWS account auth/creds not configured
skip_remote_tests: bool = os.getenv("QBRAID_RUN_REMOTE_TESTS") is None
REASON = "QBRAID_RUN_REMOTE_TESTS not set (requires configuration of AWS storage)"


def gen_rand_str(length: int) -> str:
    """Returns a random alphanumeric string of specified length."""
    chars = string.ascii_letters + string.digits
    return "".join(random.choices(chars, k=length))


def get_braket_most_busy():
    """Return the most busy device for testing purposes."""
    provider = BraketProvider()
    braket_devices = provider.get_devices(
        types=["QPU"], statuses=["ONLINE"], provider_names=["Rigetti", "IonQ", "Oxford"]
    )
    qbraid_devices = [device_wrapper(device.arn) for device in braket_devices]
    test_device = None
    max_queued = 0
    for device in qbraid_devices:
        jobs_queued = device.queue_depth()
        if jobs_queued is not None and jobs_queued > max_queued:
            max_queued = jobs_queued
            test_device = device
    return test_device


@pytest.mark.skipif(skip_remote_tests, reason=REASON)
def test_braket_queue_visibility():
    """Test methods that check Braket device/job queue."""
    circuit = Circuit().h(0).cnot(0, 1)
    device = get_braket_most_busy()
    if device is None:
        pytest.skip("No devices available for testing")
    else:
        job = device.run(circuit, shots=10)
        queue_position = job.queue_position()
        job.cancel()
        assert isinstance(queue_position, int)


@pytest.mark.skipif(skip_remote_tests, reason=REASON)
@pytest.mark.parametrize(
    "company,region", [("rigetti", "us-west-1"), ("ionq", "us-east-1"), ("oqc", "eu-west-2")]
)
def test_get_region_name(company, region):
    """Test getting the AWS region name."""
    provider = BraketProvider()
    fake_arn = f"arn:aws:braket:::device/qpu/{company}/device"
    assert provider._get_region_name(fake_arn) == region


@pytest.mark.skipif(skip_remote_tests, reason=REASON)
def test_job_wrapper_type():
    """Test that job wrapper creates object of original job type"""
    device = device_wrapper("aws_dm_sim")
    circuit = Circuit().h(0).cnot(0, 1)
    job_0 = device.run(circuit, shots=10)
    job_1 = job_wrapper(job_0.id)
    assert isinstance(job_0, type(job_1))
    assert job_0.vendor_job_id == job_1.metadata()["vendorJobId"]


@pytest.mark.skipif(skip_remote_tests, reason=REASON)
def test_is_available():
    """Test device availability function."""
    provider = BraketProvider()
    devices = provider.get_devices()
    for device in devices:
        qbraid_device = BraketDevice(aws_device=device)
        is_available_bool, _, _ = qbraid_device.availability_window()
        assert device.is_available == is_available_bool


@pytest.mark.skipif(skip_remote_tests, reason=REASON)
def test_get_quantum_tasks_by_tag():
    """Test getting tagged quantum tasks."""
    provider = BraketProvider()
    all_regions = list(provider.REGIONS)
    default_region = provider._get_default_region()
    alt_regions = [e for e in all_regions if e != default_region]
    circuit = Circuit().h(0).cnot(0, 1)
    device = provider.get_device("arn:aws:braket:::device/quantum-simulator/amazon/sv1")
    key, value1, value2 = tuple(gen_rand_str(7) for _ in range(3))
    task1 = device.run(circuit, shots=2, tags={key: value1})
    task2 = device.run(circuit, shots=2, tags={key: value2})
    assert set([task1.id, task2.id]) == set(provider.get_tasks_by_tag(key))
    assert len(provider.get_tasks_by_tag(key, values=[value1], region_names=[default_region])) == 1
    assert len(provider.get_tasks_by_tag(key, region_names=alt_regions)) == 0



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing functions to convert between OpenQASM 2 and PyTKET.

"""
from typing import TYPE_CHECKING

from pytket.qasm import circuit_from_qasm_str, circuit_to_qasm_str

if TYPE_CHECKING:
    import pytket.circuit


def qasm2_to_pytket(qasm: str) -> "pytket.circuit.Circuit":
    """Returns a pytket circuit equivalent to the input OpenQASM 2 string.

    Args:
        qasm (str): OpenQASM 2 string to convert to a pytket circuit.

    Returns:
        pytket.circuit.Circuit: PyTKET circuit object equivalent to input OpenQASM 2 string.
    """
    return circuit_from_qasm_str(qasm)


def pytket_to_qasm2(circuit: "pytket.circuit.Circuit") -> str:
    """Returns an OpenQASM 2 string equivalent to the input pytket circuit.

    Args:
        circuit (pytket.circuit.Circuit): PyTKET circuit to convert to OpenQASM 2 string.

    Returns:
        str: OpenQASM 2 string equivalent to input pytket circuit.
    """
    return circuit_to_qasm_str(circuit)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing one-step functions for converting between supported
quantum software program types.

.. currentmodule:: qbraid.transpiler.conversions

Functions
----------

.. autosummary::
   :toctree: ../stubs/

    remove_qasm_barriers
    unfold_qasm_gate_defs
    flatten_qasm_program
    decompose_qasm_qelib1

Submodules
-----------

.. autosummary::
   :toctree: ../stubs/

   braket
   cirq
   openqasm3
   pennylane
   pyquil
   pytket
   qiskit

"""
import importlib
import inspect

from .qasm_passes import flatten_qasm_program, remove_qasm_barriers, unfold_qasm_gate_defs
from .qasm_qelib1 import decompose_qasm_qelib1

# Dynamically import QPROGRAM_LIBS when needed
_qbraid = importlib.import_module("qbraid._qprogram")
_PROGRAM_LIBS = getattr(_qbraid, "_PROGRAM_LIBS", [])

# List to store the names of the imported functions
conversion_functions = []

# Base path for the sub-modules
base_path = "qbraid.transpiler.conversions."

# Iterate over the installed libraries
for lib in _PROGRAM_LIBS:
    try:
        # Dynamically import the sub-module
        sub_module = importlib.import_module(base_path + lib)

        # Extract function names from the sub-module
        function_names = [
            name
            for name in dir(sub_module)
            if callable(getattr(sub_module, name)) and inspect.isfunction(getattr(sub_module, name))
        ]

        # Add functions to the current namespace
        for name in function_names:
            globals()[name] = getattr(sub_module, name)
            conversion_functions.append(name)

    except ModuleNotFoundError:
        pass



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Tests of functions that create and operate on directed graph
used to dictate transpiler conversions.

"""
import braket.circuits
import networkx as nx
import pytest
from qiskit_braket_provider.providers.adapter import convert_qiskit_to_braket_circuit

from qbraid._qprogram import QPROGRAM_LIBS
from qbraid.transpiler.conversions import conversion_functions
from qbraid.transpiler.converter import convert_to_package
from qbraid.transpiler.edge import Conversion
from qbraid.transpiler.graph import ConversionGraph


def bound_method_str(source, target):
    """Inserts package names into string representation of bound method."""
    return f"<bound method Conversion.convert of ('{source}', '{target}')>"


def are_graphs_equal(graph1: nx.DiGraph, graph2: nx.DiGraph) -> bool:
    """Return True if two graphs are equal, False otherwise."""
    # Check if nodes are the same
    if set(graph1.nodes) != set(graph2.nodes):
        return False

    # Check if edges are the same
    if set(graph1.edges) != set(graph2.edges):
        return False

    # Check if node attributes are the same
    for node in graph1.nodes:
        if graph1.nodes[node] != graph2.nodes.get(node, {}):
            return False

    # Check if edge attributes are the same
    for edge in graph1.edges:
        if graph1.edges[edge] != graph2.edges.get(edge, {}):
            return False

    return True


@pytest.mark.parametrize("func", conversion_functions)
def test_conversion_functions_syntax(func):
    """Test that all conversion functions are named correctly."""
    source, target = func.split("_to_")
    assert source in QPROGRAM_LIBS
    assert target in QPROGRAM_LIBS
    assert source != target


def test_shortest_conversion_path():
    """Test that the shortest conversion path is found correctly."""
    G = ConversionGraph()
    shortest_path = G.find_shortest_conversion_path("qiskit", "cirq")
    top_paths = G.find_top_shortest_conversion_paths("qiskit", "cirq", top_n=3)
    assert str(shortest_path[0]) == bound_method_str("qiskit", "qasm2")
    assert str(shortest_path[1]) == bound_method_str("qasm2", "cirq")
    assert shortest_path == top_paths[0]
    assert len(top_paths) == 3 and len(top_paths[0]) <= len(top_paths[1]) <= len(top_paths[2])


def test_add_conversion():
    """
    Test the addition of a new conversion to the ConversionGraph.

    This test checks:
    1. Whether a new conversion edge is correctly added to the graph.
    2. If the graph structure with the new edge matches the expected structure.
    3. The correctness of the shortest conversion path after adding the new edge.
    """
    # Setup - preparing Conversion and ConversionGraph instances
    source, target = "cirq", "qir"
    conversion_func = lambda x: x  # pylint: disable=unnecessary-lambda-assignment
    new_edge = Conversion(source, target, conversion_func)
    initial_conversions = ConversionGraph.load_default_conversions()
    graph_with_new_edge = ConversionGraph(initial_conversions + [new_edge])
    graph_without_new_edge = ConversionGraph(initial_conversions)

    # Action - adding the new conversion edge to the graph
    graph_without_new_edge.add_conversion(new_edge)

    # Assertion 1 - Check if the new edge is added to the graph
    assert graph_without_new_edge.has_edge(source, target)

    # Assertion 2 - Check if the updated graph matches the expected graph structure
    expected_graph = graph_with_new_edge
    updated_graph = graph_without_new_edge
    assert are_graphs_equal(updated_graph, expected_graph)

    # Assertion 3 - Verify the shortest path after adding the new edge
    expected_shortest_path = [
        bound_method_str("qiskit", "qasm2"),
        bound_method_str("qasm2", "cirq"),
        bound_method_str("cirq", "qir"),
    ]
    actual_shortest_path = graph_without_new_edge.find_shortest_conversion_path("qiskit", "qir")
    assert [str(bound_method) for bound_method in actual_shortest_path] == expected_shortest_path


@pytest.mark.parametrize("bell_circuit", ["qiskit"], indirect=True)
def test_initialize_new_conversion(bell_circuit):
    """Test initializing the conversion graph with a new conversion"""
    qiskit_circuit, _ = bell_circuit
    conversions = [
        Conversion(
            "qiskit",
            "braket",
            convert_qiskit_to_braket_circuit,
        )
    ]
    graph = ConversionGraph(conversions)
    assert len(graph.edges) == 1
    braket_circuit = convert_to_package(qiskit_circuit, "braket", conversion_graph=graph)
    assert isinstance(braket_circuit, braket.circuits.Circuit)


@pytest.mark.parametrize("bell_circuit", ["qiskit"], indirect=True)
def test_overwrite_new_conversion(bell_circuit):
    """Test dynamically adding a new conversion  the conversion graph"""
    qiskit_circuit, _ = bell_circuit
    conversions = [Conversion("qiskit", "braket", lambda x: x)]
    graph = ConversionGraph(conversions)
    assert len(graph.edges) == 1
    edge = Conversion("qiskit", "braket", convert_qiskit_to_braket_circuit)
    graph.add_conversion(edge, overwrite=True)
    assert len(graph.edges) == 1
    braket_circuit = convert_to_package(qiskit_circuit, "braket", conversion_graph=graph)
    assert isinstance(braket_circuit, braket.circuits.Circuit)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining exceptions for errors raised while processing a device.

"""
from qbraid.exceptions import QbraidError


class ProgramValidationError(QbraidError):
    """Base class for errors raised while validating a quantum program."""


class QbraidRuntimeError(QbraidError):
    """Base class for errors raised while submitting a quantum job."""


class JobError(QbraidError):
    """Base class for errors raised by Jobs."""


class JobStateError(JobError):
    """Class for errors raised due to the state of a quantum job"""



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Benchmarking transpiler QAPS metric (Quantum Area Per Second)

"""
from time import time

import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import zscore

from qbraid import circuit_wrapper
from qbraid.interface.random import random_circuit

# Step 1: Collecting the data
quantum_area_data = {}
values = list(range(25, 201, 25))
depth = 5

for num_qubits in values:
    quantum_area = num_qubits * depth

    if quantum_area not in quantum_area_data:
        quantum_area_data[quantum_area] = []

    cirq_circuit = random_circuit("cirq", num_qubits=num_qubits, depth=depth)
    qbraid_circuit = circuit_wrapper(cirq_circuit)

    start = time()
    qbraid_circuit.transpile("qiskit")
    stop = time()

    quantum_area_data[quantum_area].append(stop - start)

# Step 2: Calculating average execution times and performing linear regression
quantum_areas = []
avg_execution_times = []

for quantum_area, execution_times in quantum_area_data.items():
    avg_execution_time = np.mean(execution_times)
    quantum_areas.append(quantum_area)
    avg_execution_times.append(avg_execution_time)

x = np.array(quantum_areas)
y = np.array(avg_execution_times)
coefficients = np.polyfit(x, y, 1)

polynomial = np.poly1d(coefficients)
ys = polynomial(x)
z_scores = zscore(y - ys)

filtered_indices = np.where(np.abs(z_scores) < 2)
filtered_x = x[filtered_indices]
filtered_y = y[filtered_indices]

print(f"Linear Regression Equation: y = {coefficients[0]:.5f}x + {coefficients[1]:.5f}")
print(f"Quantum Area Per Second (QAPS): {1/coefficients[0]:.5f}")

plt.figure()
plt.scatter(filtered_x, filtered_y, label="Data Points")
plt.plot(x, ys, color="red", label="Linear Regression")
plt.xlabel("Quantum Area (num_qubits * depth)")
plt.ylabel("Average Execution Time (s)")
plt.title("Average Execution Time vs Quantum Area")
plt.grid(True)
plt.legend()
plt.show()



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for equivalence of interfacing quantum programs

"""
import numpy as np
import pytest
from braket.circuits import Circuit as BKCircuit
from cirq import Circuit, LineQubit, X, Y, Z

from qbraid import circuit_wrapper
from qbraid.inspector import get_program_type
from qbraid.interface.circuit_equality import circuits_allclose
from qbraid.interface.random.random import random_circuit, random_unitary_matrix

from ..fixtures import packages_bell, packages_shared15


def pair_packages(packages):
    """Return a list of tuples of packages to compare"""
    return [(packages[i], packages[i + 1]) for i in range(len(packages) - 1)]


bell_pairs = pair_packages(packages_bell)
shared15_pairs = pair_packages(packages_shared15)


@pytest.mark.parametrize("two_bell_circuits", bell_pairs, indirect=True)
def test_compare_bell_circuits(two_bell_circuits):
    """Test unitary equivalance of bell circuits across packages for
    testing baseline."""
    circuit1, circuit2, _, _ = two_bell_circuits
    assert circuits_allclose(circuit1, circuit2, strict_gphase=True)


@pytest.mark.parametrize("two_shared15_circuits", shared15_pairs, indirect=True)
def test_compare_shared15_circuits(two_shared15_circuits):
    """Test unitary equivalance of shared15 circuits across packages for
    testing baseline."""
    circuit1, circuit2, package1, package2 = two_shared15_circuits
    strict_gphase = not (
        "pyquil" in {package1, package2} or {package1, package2} == {"qasm2", "qasm3"}
    )
    assert circuits_allclose(circuit1, circuit2, strict_gphase=strict_gphase)


@pytest.mark.parametrize("package", ["braket", "cirq", "qiskit"])
def test_random(package):
    """Test generating random circuits"""
    program = random_circuit(package)
    assert get_program_type(program) == package


def test_collapse_empty_braket_cirq():
    """Test unitary equivalance after converting to contiguous qubits"""
    # pylint: disable=no-member
    braket_circuit = BKCircuit()
    braket_circuit.x(0)
    braket_circuit.y(2)
    braket_circuit.z(4)
    # pylint: enable=no-member
    assert braket_circuit.qubit_count == 3

    cirq_circuit = Circuit()
    q0 = LineQubit(0)
    q2 = LineQubit(2)
    q4 = LineQubit(4)
    cirq_circuit.append(X(q0))
    cirq_circuit.append(Y(q2))
    cirq_circuit.append(Z(q4))
    assert len(cirq_circuit.all_qubits()) == 3

    assert circuits_allclose(braket_circuit, cirq_circuit, strict_gphase=True)

    qprogram = circuit_wrapper(braket_circuit)
    qprogram.remove_idle_qubits()
    braket_compat_circuit = qprogram.program
    assert braket_compat_circuit.qubit_count == 3

    qprogram = circuit_wrapper(cirq_circuit)
    qprogram.remove_idle_qubits()
    cirq_compat_circuit = qprogram.program
    assert circuits_allclose(braket_compat_circuit, cirq_compat_circuit, strict_gphase=True)

    qprogram = circuit_wrapper(cirq_circuit)
    qprogram.populate_idle_qubits()
    cirq_expanded_circuit = qprogram.program
    assert len(cirq_expanded_circuit.all_qubits()) == 5


def test_random_unitary():
    """Test generating random unitary"""
    matrix = random_unitary_matrix(2)
    assert np.allclose(matrix @ matrix.conj().T, np.eye(2))



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for defining custom conversions

"""
from unittest.mock import Mock

import cirq
import pytest

from qbraid.exceptions import PackageValueError
from qbraid.interface.random import random_circuit
from qbraid.transpiler.conversions.braket import braket_to_cirq
from qbraid.transpiler.edge import Conversion


def test_raise_for_unsupported_program_input():
    """Test that an exception is raised for an unsupported program input."""
    conversion = Conversion("braket", "cirq", braket_to_cirq)
    circuit = Mock()
    with pytest.raises(PackageValueError):
        conversion.convert(circuit)


def test_raise_for_source_program_input_mismatch():
    """Test that an exception is raised for a mismatch between the source package and input."""
    source, target, other = "braket", "cirq", "qiskit"
    conversion = Conversion(source, target, braket_to_cirq)
    circuit = random_circuit(other)
    with pytest.raises(ValueError):
        conversion.convert(circuit)


def test_braket_cirq_custom_conversion():
    """Test that a Braket to Cirq conversion is successful."""
    source, target = "braket", "cirq"
    conversion = Conversion(source, target, braket_to_cirq)
    braket_circuit = random_circuit(source)
    cirq_circuit = conversion.convert(braket_circuit)
    assert isinstance(cirq_circuit, cirq.Circuit)


def test_conversion_repr():
    """Test that the conversion repr is correct."""
    source, target = "braket", "cirq"
    conversion = Conversion(source, target, braket_to_cirq)
    assert repr(conversion) == f"('{source}', '{target}')"



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module providing tools to map, analyze, and visualize conversion paths between different
quantum programs available through the qbraid.transpiler using directed graphs.

"""
from importlib import import_module
from typing import List, Optional

import networkx as nx

from qbraid.transpiler.conversions import conversion_functions
from qbraid.transpiler.edge import Conversion
from qbraid.transpiler.exceptions import ConversionPathNotFoundError


class ConversionGraph(nx.DiGraph):
    """
    Class for coordinating conversions between different quantum software programs

    """

    def __init__(self, conversions: Optional[List[Conversion]] = None):
        """
        Initialize a ConversionGraph instance.

        Args:
            conversions (optional, List[Conversion]): List of conversion edges. If None, default
                                                          conversion edges are created.
        """
        super().__init__()
        self._conversions = conversions or self.load_default_conversions()
        self.create_conversion_graph()

    @staticmethod
    def load_default_conversions() -> List[Conversion]:
        """
        Create a list of default conversion nodes using predefined conversion functions.

        Returns:
            List[Conversion]: List of default conversion edges.
        """
        transpiler = import_module("qbraid.transpiler.conversions")
        return [
            Conversion(*conversion.split("_to_"), getattr(transpiler, conversion))
            for conversion in conversion_functions
        ]

    def create_conversion_graph(self) -> None:
        """
        Create a directed graph from a list of conversion functions.

        Returns:
            None
        """
        for edge in self._conversions:
            self.add_edge(edge.source, edge.target, native=edge.native, func=edge.convert)

    def conversions(self) -> List[Conversion]:
        """
        Get the list of conversion edges.

        Returns:
            List[Conversion]: The conversion edges of the graph.
        """
        return self._conversions

    def add_conversion(self, edge: Conversion, overwrite: bool = False) -> None:
        """
        Add a new conversion function as an edge in the graph.

        Args:
            edge (qbraid.interface.Conversion): The conversion edge to add to the graph.
            overwrite (optional, bool): If True, overwrites an existing conversion.
                                        Defaults to False.

        Raises:
            ValueError: If the conversion already exists and overwrite_existing is False.
        """
        source, target = edge.source, edge.target

        if self.has_edge(source, target) and not overwrite:
            raise ValueError(
                f"Conversion from {source} to {target} already exists. "
                "Set overwrite=True to overwrite."
            )

        for old_edge in self._conversions:
            if old_edge.source == source and old_edge.target == target:
                self._conversions.remove(old_edge)
                break

        self._conversions.append(edge)
        self.add_edge(source, target, native=edge.native, func=edge.convert)

    def find_shortest_conversion_path(self, source: str, target: str) -> List[str]:
        """
        Find the shortest conversion path between two nodes in a graph.

        Args:
            source (str): The starting node for the path.
            target (str): The target node for the path.

        Returns:
            list of str: The shortest conversion path as a list of step strings.

        Raises:
            ValueError: If no path is found between source and target.
        """
        try:
            path = nx.shortest_path(self, source, target)
            return [self[path[i]][path[i + 1]]["func"] for i in range(len(path) - 1)]
        except nx.NetworkXNoPath as err:
            raise ConversionPathNotFoundError(source, target) from err

    def find_top_shortest_conversion_paths(
        self, source: str, target: str, top_n: int = 3
    ) -> List[List[str]]:
        """
        Find the top shortest conversion paths between two nodes in a graph.

        Args:
            source (str): The starting node for the path.
            target (str): The target node for the path.
            top_n (int): Number of top shortest paths to find.

        Returns:
            list of list of str: The top shortest conversion paths.

        Raises:
            ValueError: If no path is found between source and target.
        """
        try:
            all_paths = list(nx.all_simple_paths(self, source, target))
            sorted_paths = sorted(all_paths, key=len)[:top_n]
            return [
                [self[path[i]][path[i + 1]]["func"] for i in range(len(path) - 1)]
                for path in sorted_paths
            ]
        except nx.NetworkXNoPath as err:
            raise ConversionPathNotFoundError(source, target) from err

    def has_path(self, source: str, target: str) -> bool:
        """
        Check if a conversion between two languages is supported.

        Args:
            source (str): The source language.
            target (str): The target language.

        Returns:
            bool: True if the conversion is supported, False otherwise.
        """
        return nx.has_path(self, source, target)

    def reset(self, conversions: Optional[List[Conversion]] = None) -> None:
        """
        Reset the graph to its default state.

        Returns:
            None
        """
        self.clear()
        self._conversions = conversions or self.load_default_conversions()
        self.create_conversion_graph()

    def plot(self, **kwargs):
        """
        Plot the conversion graph.

        Args:
            **kwargs: Keyword arguments for the plot function.

        Returns:
            None
        """
        # pylint: disable=import-outside-toplevel
        from qbraid.visualization.plot_conversions import plot_conversion_graph

        plot_conversion_graph(self, **kwargs)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for qbraid.programs.pytket.PytketCircuit

"""
import numpy as np
import pytest
from pytket.circuit import Circuit

from qbraid.programs.pytket import PytketCircuit


@pytest.mark.parametrize(
    "input_circuit, expected_circuit",
    [
        (Circuit(2).CX(0, 1), Circuit(2).CX(1, 0)),
        (Circuit(3).CCX(0, 1, 2), Circuit(3).CCX(2, 1, 0)),
    ],
)
def test_reverse_qubit_order(input_circuit, expected_circuit):
    """Test reversing qubit ordering of pytket circuit."""
    qprogram = PytketCircuit(input_circuit)
    qprogram.reverse_qubit_order()
    result_circuit = qprogram.program
    assert result_circuit.get_commands() == expected_circuit.get_commands()


def test_remove_idle_qubits_pytket():
    """Test wires with no operations from pytket circuit"""
    circuit = Circuit(3)
    circuit.H(0)
    circuit.CX(0, 1)
    qprogram = PytketCircuit(circuit)
    qprogram.remove_idle_qubits()
    contig_circuit = qprogram.program
    assert contig_circuit.n_qubits == 2


@pytest.mark.parametrize("flat", [True, False])
@pytest.mark.parametrize("list_type", [True, False])
def test_gate_to_matrix_pytket(flat, list_type):
    """Test converting pytket gates to matrix"""
    c = Circuit(10, 2, name="example")
    c.CU1(np.pi / 2, 2, 3)

    c_unitary = PytketCircuit.gate_to_matrix(
        gates=c.get_commands()[0] if list_type else c.get_commands(), flat=flat
    )
    if flat:
        assert c_unitary.shape[0] == 2**2
    else:
        assert c_unitary.shape[0] == 2**4



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing top-level qbraid wrapper functionality. Each of these
functions utilize entrypoints via ``pkg_resources``.

"""
import openqasm3
import pkg_resources

from ._qprogram import QPROGRAM
from .exceptions import QbraidError
from .inspector import get_program_type


def _get_entrypoints(group: str):
    """Returns a dictionary mapping each entry of ``group`` to its loadable entrypoint."""
    return {entry.name: entry for entry in pkg_resources.iter_entry_points(group)}


def circuit_wrapper(program: QPROGRAM):
    """Apply qbraid quantum program wrapper to a supported quantum program.

    This function is used to create a qBraid :class:`~qbraid.transpiler.QuantumProgram`
    object, which can then be transpiled to any supported quantum circuit-building package.
    The input quantum circuit object must be an instance of a circuit object derived from a
    supported package.

    .. code-block:: python

        cirq_circuit = cirq.Circuit()
        q0, q1, q2 = [cirq.LineQubit(i) for i in range(3)]
        ...

    Please refer to the documentation of the individual qbraid circuit wrapper objects to see
    any additional arguments that might be supported.

    Args:
        circuit (:data:`~qbraid.QPROGRAM`): A supported quantum circuit / program object

    Returns:
        :class:`~qbraid.transpiler.QuantumProgram`: A wrapped quantum circuit-like object

    Raises:
        :class:`~qbraid.QbraidError`: If the input circuit is not a supported quantum program.

    """
    if isinstance(program, openqasm3.ast.Program):
        program = openqasm3.dumps(program)

    try:
        package = get_program_type(program)
    except QbraidError as err:
        raise QbraidError(
            f"Error applying circuit wrapper to quantum program \
            of type {type(program)}"
        ) from err

    transpiler_entrypoints = _get_entrypoints("qbraid.programs")
    if package not in transpiler_entrypoints:
        raise QbraidError(
            f"Error applying circuit wrapper to quantum program of type {type(program)}"
        )

    circuit_wrapper_class = transpiler_entrypoints[package].load()
    return circuit_wrapper_class(program)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Fixtures imported/defined in this file can be used by any test in this directory
without needing to import them (pytest will automatically discover them).

"""
# pylint: disable=unused-import

from .fixtures import (
    bell_circuit,
    bell_unitary,
    packages_bell,
    packages_shared15,
    shared15_circuit,
    shared15_unitary,
    two_bell_circuits,
    two_shared15_circuits,
)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.
"""
Module for plotting qBraid transpiler quantum program conversion graphs.

"""
from typing import TYPE_CHECKING, Dict, Optional

import matplotlib.pyplot as plt
import networkx as nx

from qbraid._qprogram import QPROGRAM_LIBS

if TYPE_CHECKING:
    import qbraid.transpiler


def plot_conversion_graph(  # pylint: disable=too-many-arguments
    graph: "qbraid.transpiler.ConversionGraph",
    title: Optional[str] = "qBraid Quantum Program Conversion Graph",
    legend: bool = False,
    seed: Optional[int] = None,
    node_size: int = 1200,
    min_target_margin: int = 18,
    show: bool = True,
    save_path: Optional[str] = None,
    colors: Optional[Dict[str, str]] = None,
) -> None:
    """
    Plot the conversion graph using matplotlib. The graph is displayed using node
    and edge color conventions, with options for a title, legend, and figure saving.

    Args:
        graph (qbraid.interface.ConversionGraph): The directed conversion graph to be plotted.
        title (Optional[str]): Title of the plot. Defaults to
                               'qBraid Quantum Program Conversion Graph'.
        legend (bool): If True, display a legend on the graph. Defaults to False.
        seed (Optional[int]): Seed for the node layout algorithm. Useful for consistent positioning.
                              Defaults to None.
        node_size (int): Size of the nodes. Defaults to 1200.
        min_target_margin (int): Minimum target margin for edges. Defaults to 18.
        show (bool): If True, display the figure. Defaults to True.
        save_path (Optional[str]): Path to save the figure. If None, the figure is not saved.
                                   Defaults to None.
        colors (Optional[Dict[str, str]]): Dictionary for node and edge colors. Expected keys are
            'qbraid_node', 'external_node', 'qbraid_edge', 'external_edge'. Defaults to None.

    Returns:
        None
    """
    # Set default colors if not provided
    if colors is None:
        colors = {
            "qbraid_node": "lightblue",
            "external_node": "lightgray",
            "qbraid_edge": "gray",
            "external_edge": "blue",
        }

    # Extract colors and apply them in the drawing
    ncolors = [
        colors["qbraid_node"] if node in QPROGRAM_LIBS else colors["external_node"]
        for node in graph.nodes()
    ]

    # Create a dictionary for quick lookup of conversions by their source and target
    conversion_dict = {
        (conversion.source, conversion.target): conversion for conversion in graph.conversions()
    }
    conversions_ordered = [
        conversion_dict[(edge[0], edge[1])]
        for edge in graph.edges()
        if (edge[0], edge[1]) in conversion_dict
    ]
    ecolors = [
        (
            colors["qbraid_edge"]
            if graph[edge.source][edge.target]["native"]
            else colors["external_edge"]
        )
        for edge in conversions_ordered
    ]

    pos = nx.spring_layout(graph, seed=seed)  # good seeds: 123, 134
    nx.draw_networkx_nodes(graph, pos, node_color=ncolors, node_size=node_size)
    nx.draw_networkx_edges(graph, pos, edge_color=ecolors, min_target_margin=min_target_margin)
    nx.draw_networkx_labels(graph, pos)

    if title:
        plt.title(title)
    plt.axis("off")

    if legend:
        # Create legend elements using a loop
        legend_info = [
            ("qBraid - Node", "o", colors["qbraid_node"], None),
            ("External - Node", "o", colors["external_node"], None),
            ("qBraid - Edge", None, colors["qbraid_edge"], "-"),
            ("External - Edge", None, colors["external_edge"], "-"),
        ]
        legend_elements = [
            plt.Line2D(
                [0],
                [0],
                marker=marker,
                color="w" if marker else color,
                label=label,
                markersize=10 if marker else None,
                markerfacecolor=color if marker else None,
                linestyle=linestyle,
                linewidth=2 if linestyle else None,
            )
            for label, marker, color, linestyle in legend_info
        ]
        plt.legend(handles=legend_elements, loc="best")

    if show:
        plt.show()

    if save_path:
        plt.savefig(save_path)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for BraketDevice class

"""
import os
from datetime import datetime

import pytest
from braket.aws.aws_device import AwsDevice
from braket.tasks.quantum_task import QuantumTask as AwsQuantumTask
from qiskit import QuantumCircuit as QiskitCircuit

from qbraid import QbraidError, device_wrapper
from qbraid.interface import random_circuit
from qbraid.providers import QuantumJob
from qbraid.providers.aws import BraketDevice, BraketQuantumTask
from qbraid.providers.exceptions import ProgramValidationError

from .fixtures import braket_circuit, cirq_circuit, device_wrapper_inputs, qiskit_circuit

# Skip tests if IBM/AWS account auth/creds not configured
skip_remote_tests: bool = os.getenv("QBRAID_RUN_REMOTE_TESTS") is None
REASON = "QBRAID_RUN_REMOTE_TESTS not set (requires configuration of AWS storage)"
pytestmark = pytest.mark.skipif(skip_remote_tests, reason=REASON)

### FIXTURES ###

inputs_braket_dw = [] if skip_remote_tests else device_wrapper_inputs("AWS")

circuits_braket_run = [
    braket_circuit(),
    cirq_circuit(False),
    qiskit_circuit(False),
]  # circuits w/out measurement operation

inputs_braket_run = ["aws_sv_sim"]

### TESTS ###


def test_device_wrapper_id_error():
    """Test raising device wrapper error due to invalid device ID."""
    with pytest.raises(QbraidError):
        device_wrapper("Id123")


@pytest.mark.parametrize("device_id", inputs_braket_dw)
def test_device_wrapper_aws_from_api(device_id):
    """Test device wrapper for ids of devices available through Amazon Braket."""
    qbraid_device = device_wrapper(device_id)
    vendor_device = qbraid_device._device
    assert isinstance(qbraid_device, BraketDevice)
    assert isinstance(vendor_device, AwsDevice)


def test_device_wrapper_from_braket_arn():
    """Test creating device wrapper from Amazon Braket device ARN."""
    aws_device_arn = "arn:aws:braket:::device/quantum-simulator/amazon/sv1"
    qbraid_device = device_wrapper(aws_device_arn)
    vendor_device = qbraid_device._device
    assert isinstance(qbraid_device, BraketDevice)
    assert isinstance(vendor_device, AwsDevice)


def test_device_wrapper_properties():
    """Test extracting properties from BraketDevice"""
    device_id = "aws_oqc_lucy"
    wrapper = device_wrapper(device_id)
    assert wrapper.provider == "Oxford"
    assert wrapper.name == "Lucy"
    assert str(wrapper) == "AWS Oxford Lucy device wrapper"
    assert repr(wrapper) == "<BraketDevice(Oxford:'Lucy')>"


def test_queue_depth():
    """Test getting queue_depth BraketDevice"""
    aws_device = device_wrapper("aws_sv_sim")
    assert isinstance(aws_device.queue_depth(), int)


@pytest.mark.parametrize("circuit", circuits_braket_run)
@pytest.mark.parametrize("device_id", inputs_braket_run)
def test_run_braket_device_wrapper(device_id, circuit):
    """Test run method of wrapped Braket devices"""
    qbraid_device = device_wrapper(device_id)
    qbraid_job = qbraid_device.run(circuit, shots=10)
    vendor_job = qbraid_job._job
    try:
        qbraid_job.cancel()
    except Exception:  # pylint: disable=broad-exception-caught
        pass
    assert isinstance(qbraid_job, BraketQuantumTask)
    assert isinstance(vendor_job, AwsQuantumTask)


def test_run_batch_braket_device_wrapper():
    """Test run_batch method of wrapped Braket devices"""
    qbraid_device = device_wrapper("aws_sv_sim")
    qbraid_job_list = qbraid_device.run_batch(circuits_braket_run, shots=10)
    qbraid_job = qbraid_job_list[0]
    for job in qbraid_job_list:
        try:
            job.cancel()
        except Exception:  # pylint: disable=broad-exception-caught
            pass
    assert len(qbraid_job_list) == len(circuits_braket_run)
    assert isinstance(qbraid_job, BraketQuantumTask)


def test_circuit_too_many_qubits():
    """Test that run method raises exception when input circuit
    num qubits exceeds that of wrapped AWS device."""
    device = device_wrapper("aws_ionq_harmony")
    num_qubits = device.num_qubits + 10
    circuit = QiskitCircuit(num_qubits)
    circuit.h([0, 1])
    circuit.cx(0, num_qubits - 1)
    with pytest.raises(ProgramValidationError):
        device.run(circuit)


def test_device_num_qubits():
    """Test device wrapper num qubits method"""
    five_qubit_device = device_wrapper("aws_ionq_harmony")
    assert five_qubit_device.num_qubits == 11


def test_wait_for_final_state():
    """Test function that returns after job is complete"""
    device = device_wrapper("aws_dm_sim")
    circuit = random_circuit("qiskit")
    job = device.run(circuit, shots=10)
    job.wait_for_final_state()
    status = job.status()
    assert QuantumJob.status_final(status)


def test_aws_device_available():
    """Test BraketDeviceWrapper avaliable output identical"""
    device = device_wrapper("aws_dm_sim")
    is_available_bool, is_available_time, iso_str = device.availability_window()
    assert is_available_bool == device._device.is_available
    assert len(is_available_time.split(":")) == 3
    assert isinstance(iso_str, str)
    try:
        datetime.strptime(iso_str, "%Y-%m-%dT%H:%M:%SZ")
    except ValueError:
        pytest.fail("iso_str not in expected format")



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for generate random quantum circuits used for testing

"""
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple

import numpy as np

from qbraid._qprogram import QPROGRAM, QPROGRAM_LIBS
from qbraid.exceptions import PackageValueError
from qbraid.transpiler.converter import convert_to_package

QROGRAM_TEST_TYPE = Tuple[Dict[str, Callable[[Any], QPROGRAM]], np.ndarray]

if TYPE_CHECKING:
    import qbraid


def random_circuit(
    package: str, num_qubits: Optional[int] = None, depth: Optional[int] = None, **kwargs
) -> "qbraid.QPROGRAM":
    """Generate random circuit of arbitrary size and form.

    Args:
        package: qBraid supported software package
        num_qubits: Number of quantum wires. If not provided, set randomly in range [2,4].
        depth: Layers of operations (i.e. critical path length)
            If not provided, set randomly in range [2,4].

    Raises:
        PackageValueError: if ``package`` is not supported
        QbraidError: when invalid random circuit options given

    Returns:
        :data:`~qbraid.QPROGRAM`: randomly generated quantum circuit/program

    """
    if package not in QPROGRAM_LIBS:
        raise PackageValueError(package)

    num_qubits = np.random.randint(1, 4) if num_qubits is None else num_qubits
    depth = np.random.randint(1, 4) if depth is None else depth

    # pylint: disable=import-outside-toplevel
    if package == "qasm3":
        from qbraid.interface.random.qasm3_random import _qasm3_random

        rand_circuit = _qasm3_random(num_qubits, depth, **kwargs)
    elif package == "qiskit":
        from qbraid.interface.random.qiskit_random import _qiskit_random

        rand_circuit = _qiskit_random(num_qubits, depth, **kwargs)
    else:
        from qbraid.interface.random.cirq_random import _cirq_random

        rand_circuit = _cirq_random(num_qubits, depth, **kwargs)

        if package != "cirq":
            rand_circuit = convert_to_package(rand_circuit, package)

    return rand_circuit


def random_unitary_matrix(dim: int) -> np.ndarray:
    """Create a random (complex) unitary matrix of order `dim`

    Args:
        dim: integer square matrix dimension

    Returns:
        random unitary matrix of shape dim x dim
    """
    # Create a random complex matrix of size dim x dim
    matrix = np.random.randn(dim, dim) + 1j * np.random.randn(dim, dim)
    # Use the QR decomposition to get a random unitary matrix
    unitary, _ = np.linalg.qr(matrix)
    return unitary



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for conversions between Cirq circuits and Qiskit circuits.

"""
import numpy as np
import pytest
from cirq import Circuit, LineQubit, ops, testing

from qbraid.interface import circuits_allclose
from qbraid.transpiler.converter import convert_to_package
from qbraid.transpiler.exceptions import CircuitConversionError


def test_bell_state_to_qiskit():
    """Tests cirq.Circuit --> qiskit.QuantumCircuit --> cirq.Circuit
    with a Bell state circuit."""
    qreg = LineQubit.range(2)
    cirq_circuit = Circuit([ops.H.on(qreg[0]), ops.CNOT.on(qreg[0], qreg[1])])
    qiskit_circuit = convert_to_package(cirq_circuit, "qiskit")
    assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=True)


@pytest.mark.parametrize("num_qubits", [1, 2, 3, 4, 5])
def test_random_circuit_to_qiskit(num_qubits):
    """Tests converting random Cirq circuits to Qiskit circuits."""
    for _ in range(10):
        cirq_circuit = testing.random_circuit(
            qubits=num_qubits,
            n_moments=np.random.randint(1, 6),
            op_density=1,
            random_state=np.random.randint(1, 10),
        )
        qiskit_circuit = convert_to_package(cirq_circuit, "qiskit")
        assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=True)


def test_raise_error():
    """Tests raising error for unsupported gates."""
    with pytest.raises(CircuitConversionError):
        probs = np.random.uniform(low=0, high=0.5)
        cirq_circuit = Circuit(ops.PhaseDampingChannel(probs).on(*LineQubit.range(1)))
        convert_to_package(cirq_circuit, "qiskit")



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Benchmarking tests for Cirq conversions

"""
import string

import cirq
import numpy as np
import pytest
import scipy

import qbraid


def generate_params(varnames, seed=0):
    """Generate random parameters to help construct Cirq test gates"""
    np.random.seed(seed)
    params = {}
    # Generating angles
    for ra in ["rads", "theta", "phi"]:
        if ra in varnames:
            params[ra] = np.random.rand() * 2 * np.pi
    # Generating probabilities
    for prob in ["gamma", "p", "probability"]:
        if prob in varnames:
            params[prob] = np.random.rand()

    # exponents
    for exp in ["exponent", "x_exponent", "z_exponent", "phase_exponent", "axis_phase_exponent"]:
        if exp in varnames:
            params[exp] = np.random.rand() * 10

    if "num_qubits" in varnames:
        params["num_qubits"] = np.random.randint(1, 7)

    if "sub_gate" in varnames:
        params["sub_gate"] = np.random.choice([cirq.X, cirq.Y, cirq.Z, cirq.S, cirq.T])

    if "matrix" in varnames:
        n = np.random.randint(1, 4)
        params["matrix"] = scipy.stats.unitary_group.rvs(2**n)

    return params


def get_cirq_gates():
    """Construct a dictionary of all Cirq gates and assign random parameters as applicable"""
    qubits = cirq.LineQubit.range(7)

    cirq_gate_dict = {
        attr: None
        for attr in dir(cirq.ops)
        if attr[0] in string.ascii_uppercase
        if (
            isinstance(
                getattr(cirq.ops, attr), (cirq.Gate, cirq.value.abc_alt.ABCMetaImplementAnyOneOf)
            )
        )
    }
    for gate in cirq_gate_dict:
        try:
            params = generate_params(getattr(cirq.ops, gate).__init__.__code__.co_varnames)
        except Exception:  # pylint: disable=broad-exception-caught
            continue

        if not isinstance(getattr(cirq.ops, gate), cirq.Gate):
            try:
                cirq_gate_dict[gate] = getattr(cirq.ops, gate)(**params)
            except Exception:  # pylint: disable=broad-exception-caught
                continue

    for gate in cirq_gate_dict:
        if cirq_gate_dict.get(gate):
            cirq_gate_dict[gate] = cirq_gate_dict[gate](
                *qubits[: cirq_gate_dict[gate].num_qubits()]
            )
        else:
            try:
                cirq_gate_dict[gate] = getattr(cirq.ops, gate)(
                    *qubits[: getattr(cirq.ops, gate).num_qubits()]
                )
            except Exception:  # pylint: disable=broad-exception-caught
                continue

    return {k: v for k, v in cirq_gate_dict.items() if v is not None}


#############
### TESTS ###
#############

TARGETS = [("braket", 0.85), ("pyquil", 0.74), ("pytket", 0.85), ("qiskit", 0.85)]
cirq_gates = get_cirq_gates()


def convert_from_cirq_to_x(target, gate_name):
    """Construct a Cirq circuit with the given gate, transpile it to
    target program type, and check equivalence.
    """
    gate = cirq_gates[gate_name]
    source_circuit = cirq.Circuit()
    source_circuit.append(gate)
    target_circuit = qbraid.circuit_wrapper(source_circuit).transpile(target)
    assert qbraid.interface.circuits_allclose(source_circuit, target_circuit, strict_gphase=False)


@pytest.mark.parametrize(("target", "baseline"), TARGETS)
def test_cirq_coverage(target, baseline):
    """Test converting Cirq circuits to supported target program type over
    all Cirq gates and check against baseline expecte accuracy.
    """
    ACCURACY_BASELINE = baseline
    ALLOWANCE = 0.01
    failures = {}
    for gate_name in cirq_gates:
        try:
            convert_from_cirq_to_x(target, gate_name)
        except Exception as e:  # pylint: disable=broad-exception-caught
            failures[f"{target}-{gate_name}"] = e

    total_tests = len(cirq_gates)
    nb_fails = len(failures)
    nb_passes = total_tests - nb_fails
    accuracy = float(nb_passes) / float(total_tests)

    assert accuracy >= ACCURACY_BASELINE - ALLOWANCE, (
        f"The coverage threshold was not met. {nb_fails}/{total_tests} tests failed "
        f"({nb_fails / (total_tests):.2%}) and {nb_passes}/{total_tests} passed "
        f"(expected >= {ACCURACY_BASELINE}).\nFailures: {failures.keys()}\n\n"
    )



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing top-level qbraid wrapper functionality. Each of these
functions utilize entrypoints via ``pkg_resources``.

"""
import pkg_resources

from .api import QbraidSession
from .exceptions import QbraidError
from .providers import QbraidProvider


def _get_entrypoints(group: str):
    """Returns a dictionary mapping each entry of ``group`` to its loadable entrypoint."""
    return {entry.name: entry for entry in pkg_resources.iter_entry_points(group)}


def _get_device_info(device_id: str):
    """Retrieve a device from qBraid API using device ID."""
    session = QbraidSession()
    api_endpoint = "/public/lab/get-devices"
    params_list = [{"qbraid_id": device_id}, {"objArg": device_id}]

    for params in params_list:
        device_lst = session.get(api_endpoint, params=params).json()
        if device_lst and len(device_lst) > 0:
            return device_lst[0]

    raise QbraidError(f"{device_id} is not a valid device ID.")


def device_wrapper(device_id: str):
    """Apply qbraid device wrapper to device from a supported device provider.

    Args:
        device_id: unique ID specifying a supported quantum hardware device/simulator

    Returns:
        :class:`~qbraid.providers.QuantumDevice`: A wrapped quantum device-like object

    Raises:
        :class:`~qbraid.QbraidError`: If ``device_id`` is not a valid device reference.

    """
    device_info = _get_device_info(device_id)

    try:
        device_id = device_info["objArg"]
    except KeyError as err:
        raise QbraidError(f"Device {device_id} does not have an associated objArg.") from err
    provider = QbraidProvider()
    device_obj = provider.get_device(device_id)

    devices_entrypoints = _get_entrypoints("qbraid.providers")

    vendor = device_info["vendor"].lower()
    ep = f"{vendor}.device"
    device_wrapper_class = devices_entrypoints[ep].load()
    return device_wrapper_class(device_obj)


def job_wrapper(qbraid_job_id: str):
    """Retrieve a job from qBraid API using job ID and return job wrapper object.

    Args:
        qbraid_job_id: qBraid Job ID

    Returns:
        :class:`~qbraid.providers.job.QuantumJob`: A wrapped quantum job-like object

    """
    session = QbraidSession()
    job_lst = session.post(
        "/get-user-jobs", json={"qbraidJobId": qbraid_job_id, "numResults": 1}
    ).json()

    if len(job_lst) == 0:
        job_lst = session.post(
            "/get-user-jobs", json={"_id": qbraid_job_id, "numResults": 1}
        ).json()

        if len(job_lst) == 0:
            raise QbraidError(f"{qbraid_job_id} is not a valid job ID.")

    job_data = job_lst[0]

    try:
        vendor_device_id = job_data["vendorDeviceId"]
        qbraid_device = device_wrapper(vendor_device_id)
    except (KeyError, QbraidError):
        qbraid_device = None

    status_str = job_data.get("qbraidStatus", job_data.get("status", "UNKNOWN"))
    try:
        vendor_job_id = job_data["vendorJobId"]
    except KeyError as err:
        raise QbraidError(f"Job {qbraid_job_id} does not have an associated vendorJobId.") from err
    vendor = qbraid_device.vendor.lower()
    devices_entrypoints = _get_entrypoints("qbraid.providers")
    ep = f"{vendor}.job"
    job_wrapper_class = devices_entrypoints[ep].load()
    return job_wrapper_class(
        qbraid_job_id, vendor_job_id=vendor_job_id, device=qbraid_device, status=status_str
    )



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for qbraid.programs.qasm2.OpenQasm2Program

"""

import pytest

from qbraid.programs.qasm2 import OpenQasm2Program

from ..fixtures.qasm2.circuits import (
    _read_qasm_file,
    qasm2_bell,
    qasm2_cirq_shared15,
    qasm2_shared15,
)


def test_qasm_qubits():
    """Test getting QASM qubits"""

    assert OpenQasm2Program(qasm2_bell()).qubits == ["q[0]", "q[1]"]
    assert OpenQasm2Program(qasm2_shared15()).qubits == ["q[0]", "q[1]", "q[2]", "q[3]"]


def test_qasm_num_qubits():
    """Test calculating number of qubits in qasm2 circuit"""
    assert OpenQasm2Program(qasm2_bell()).num_qubits == 2
    assert OpenQasm2Program(qasm2_shared15()).num_qubits == 4


QASM_DEPTH_DATA = [
    (qasm2_bell(), 2),
    (qasm2_shared15(), 7),
    (qasm2_cirq_shared15(), 22),
    (_read_qasm_file("qtp.qasm"), 9),
    (
        """
OPENQASM 2.0;
include "qelib1.inc";
qreg q[2];
h q[0];
barrier q;
h q[1];
        """,
        2,
    ),
    (
        """
OPENQASM 2.0;
include "qelib1.inc";
qreg q[2];
creg c0[1];
creg c1[1];
h q[0];
h q[1];
measure q[0] -> c0[0];
if(c0==0) x q[1];
        """,
        3,
    ),
    (
        """
OPENQASM 2.0;
include "qelib1.inc";
qreg q[2];
creg c0[1];
creg c1[1];
h q;
measure q[0] -> c0[0];
if(c0==0) x q[1];
if(c0==0) measure q[1] -> c1[0];
        """,
        4,
    ),
    (
        """
OPENQASM 2.0;
include "qelib1.inc";
qreg q[3];
qreg a[2];
creg c[3];
creg syn[2];
gate syndrome d1,d2,d3,a1,a2
{
cx d1,a1; cx d2,a1;
cx d2,a2; cx d3,a2;
}
x q[0];
barrier q;
syndrome q[0],q[1],q[2],a[0],a[1];
measure a -> syn;
if(syn==1) x q[0];
if(syn==2) x q[2];
if(syn==3) x q[1];
measure q -> c;
        """,
        7,
    ),
    (
        """
OPENQASM 2.0;
include "qelib1.inc";
gate majority a,b,c
{
cx c,b;
cx c,a;
ccx a,b,c;
}
gate unmaj a,b,c
{
ccx a,b,c;
cx c,a;
cx a,b;
}
qreg cin[1];
qreg a[4];
qreg b[4];
qreg cout[1];
creg ans[5];
x a[0];
x b;
majority cin[0],b[0],a[0];
majority a[0],b[1],a[1];
majority a[1],b[2],a[2];
majority a[2],b[3],a[3];
cx a[3],cout[0];
unmaj a[2],b[3],a[3];
unmaj a[1],b[2],a[2];
unmaj a[0],b[1],a[1];
unmaj cin[0],b[0],a[0];
measure b[0] -> ans[0];
measure b[1] -> ans[1];
measure b[2] -> ans[2];
measure b[3] -> ans[3];
measure cout[0] -> ans[4];
        """,
        11,
    ),
    (
        """
OPENQASM 2.0;
include "qelib1.inc";
qreg q[4];
creg c[4];
h q;
barrier q;
h q[0];
measure q[0] -> c[0];
if(c==1) u1(pi/2) q[1];
h q[1];
measure q[1] -> c[1];
if(c==1) u1(pi/4) q[2];
if(c==2) u1(pi/2) q[2];
if(c==3) u1(pi/2+pi/4) q[2];
h q[2];
measure q[2] -> c[2];
if(c==1) u1(pi/8) q[3];
if(c==2) u1(pi/4) q[3];
if(c==3) u1(pi/4+pi/8) q[3];
if(c==4) u1(pi/2) q[3];
if(c==5) u1(pi/2+pi/8) q[3];
if(c==6) u1(pi/2+pi/4) q[3];
if(c==7) u1(pi/2+pi/4+pi/8) q[3];
h q[3];
measure q[3] -> c[3];
        """,
        20,
    ),
]


@pytest.mark.parametrize("qasm_str, expected_depth", QASM_DEPTH_DATA)
def test_qasm_depth(qasm_str, expected_depth):
    """Test calculating depth of circuit represented by qasm2 string"""
    qprogram = OpenQasm2Program(qasm_str)
    assert qprogram.depth == expected_depth



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining all :mod:`~qbraid.providers` enumerated types.

"""
from enum import Enum


class DeviceType(str, Enum):
    """Class for possible device types.

    Attributes:
        SIMULATOR (str): the device is a simulator
        QPU (str): the device is a QPU
        FAKE (str): is a fake device used for testing

    """

    SIMULATOR = "SIMULATOR"
    QPU = "QPU"
    FAKE = "FAKE"


class DeviceStatus(int, Enum):
    """Class for the status of devices.

    Attributes:
        ONLINE (int): the device is online
        OFFLINE (int): the device is offline

    """

    ONLINE = 0
    OFFLINE = 1
    RETIRED = 2


class JobStatus(str, Enum):
    """Class for the status of processes (i.e. jobs / quantum tasks) resulting from any
    :meth:`~qbraid.providers.QuantumDevice.run` method.

    Attributes:
        INITIALIZING (str): job is being initialized
        QUEUED (str): job is queued
        VALIDATING (str): job is being validated
        RUNNING (str): job is actively running
        CANCELLING (str): job is being cancelled
        CANCELLED (str): job has been cancelled
        COMPLETED (str): job has successfully run
        FAILED (str): job failed / incurred error
        UNKNOWN (str): vendor-supplied job status not recognized

    """

    INITIALIZING = "job is being initialized"
    QUEUED = "job is queued"
    VALIDATING = "job is being validated"
    RUNNING = "job is actively running"
    CANCELLING = "job is being cancelled"
    CANCELLED = "job has been cancelled"
    COMPLETED = "job has successfully run"
    FAILED = "job failed / incurred error"
    UNKNOWN = "vendor-supplied job status not recognized"


JOB_FINAL = (JobStatus.COMPLETED, JobStatus.CANCELLED, JobStatus.FAILED)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.
"""
Module for testing datetime helper functions used for getting device availability windows

"""
from datetime import datetime
from unittest.mock import patch

import pytest

from qbraid.providers.aws.device import _future_utc_datetime


# Test function
@pytest.mark.parametrize(
    "hours, minutes, seconds, expected",
    [
        (1, 0, 0, "2023-01-01T01:00:00Z"),
        (0, 30, 0, "2023-01-01T00:30:00Z"),
        (0, 0, 45, "2023-01-01T00:00:45Z"),
    ],
)
def test_future_utc_datetime(hours, minutes, seconds, expected):
    """Test calculating future utc datetime"""
    with patch("qbraid.providers.aws.device.datetime") as mock_datetime:
        mock_datetime.utcnow.return_value = datetime(2023, 1, 1, 0, 0, 0)
        assert _future_utc_datetime(hours, minutes, seconds) == expected



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit test for the graph-based transpiler

"""
import braket.circuits
import pytest

from qbraid.transpiler.converter import _get_path_from_bound_methods, convert_to_package
from qbraid.transpiler.edge import Conversion
from qbraid.transpiler.exceptions import ConversionPathNotFoundError, NodeNotFoundError
from qbraid.transpiler.graph import ConversionGraph


def test_unuspported_target_package():
    """Test that an error is raised if target package is not supported."""
    with pytest.raises(NodeNotFoundError):
        convert_to_package(braket.circuits.Circuit(), "alice")


def test_get_path_from_bound_method():
    """Test formatted conversion path logging helper function."""
    source, target = "cirq", "qasm2"
    edge = Conversion(source, target, lambda x: x)
    graph = ConversionGraph([edge])
    bound_method = graph[source][target]["func"]
    bound_method_list = [bound_method]
    path = _get_path_from_bound_methods(bound_method_list)
    assert path == "cirq -> qasm2"


@pytest.mark.parametrize("bell_circuit", ["qiskit"], indirect=True)
def test_raise_no_conversion_path_found(bell_circuit):
    """Test raising exception when no conversion path is found"""
    qiskit_circuit, _ = bell_circuit
    conversions = [
        Conversion("cirq", "braket", lambda x: x),
        Conversion("cirq", "qiskit", lambda x: x),
    ]
    graph = ConversionGraph(conversions)
    with pytest.raises(ConversionPathNotFoundError):
        convert_to_package(qiskit_circuit, "braket", conversion_graph=graph)


@pytest.mark.parametrize("bell_circuit", ["qiskit"], indirect=True)
def test_raise_no_conversion_path_found_max_depth(bell_circuit):
    """Test raising exception when no conversion path is found when a conversion path
    exists but does not meet the max_depth requirement."""
    qiskit_circuit, _ = bell_circuit
    with pytest.raises(ConversionPathNotFoundError):
        convert_to_package(qiskit_circuit, "braket", max_path_depth=1)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Fixtures imported/defined in this file can be used by any test in this directory
without needing to import them (pytest will automatically discover them).

"""
import braket.circuits
import numpy as np
import pytest

from qbraid import device_wrapper
from qbraid.providers.aws import BraketProvider


@pytest.fixture
def braket_most_busy():
    """Return the most busy device for testing purposes."""
    provider = BraketProvider()
    braket_devices = provider.get_devices(
        types=["QPU"], statuses=["ONLINE"], provider_names=["Rigetti", "IonQ", "Oxford"]
    )
    qbraid_devices = [device_wrapper(device.arn) for device in braket_devices]
    qbraid_device = None
    max_queued = 0
    for device in qbraid_devices:
        jobs_queued = device.queue_depth()
        if jobs_queued is not None and jobs_queued > max_queued:
            max_queued = jobs_queued
            qbraid_device = device
    yield qbraid_device


@pytest.fixture
def aws_session():
    """Return AWS session."""
    provider = BraketProvider()
    yield provider._get_aws_session()


@pytest.fixture
def braket_circuit():
    """Low-depth, one-qubit Braket circuit to be used for testing."""
    circuit = braket.circuits.Circuit()
    circuit.h(0)
    circuit.ry(0, np.pi / 2)
    yield circuit



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for defining custom conversions

"""
import inspect
from typing import TYPE_CHECKING, Any, Callable, Union

from qbraid._qprogram import SUPPORTED_QPROGRAMS
from qbraid.inspector import get_program_type

if TYPE_CHECKING:
    import qbraid


class Conversion:
    """
    Class for defining and handling custom conversions between different quantum program packages.

    """

    def __init__(self, source: str, target: str, conversion_func: Callable):
        """
        Initialize a Conversion instance with source and target packages and a conversion function.

        Args:
            source (str): The source package from which conversion starts.
            target (str): The target package to which conversion is done.
            conversion_func (Callable): The function that performs the actual conversion.
        """
        self._source = source
        self._target = target
        self._conversion_func = conversion_func
        self._native = self._is_native()

    @property
    def source(self) -> str:
        """
        The source package of the conversion.

        Returns:
            str: The source package name.
        """
        return self._source

    @property
    def target(self) -> str:
        """
        The target package of the conversion.

        Returns:
            str: The target package name.
        """
        return self._target

    @property
    def native(self) -> bool:
        """
        True if the conversion function is native to qbraid package, False otherwise.

        Returns:
            bool: Whether the conversion function is native to qbraid package.
        """
        return self._native

    def _is_native(self) -> bool:
        """
        Check if a conversion function is native to qbraid package.

        Args:
            func (Callable): The conversion function to check.

        Returns:
            bool: True if the function is native, False otherwise.
        """
        module = inspect.getmodule(self._conversion_func)

        if module is None:
            return False

        package = module.__name__.split(".")[0]
        return package == "qbraid"

    def convert(self, program: "qbraid.QPROGRAM") -> Union["qbraid.QPROGRAM", Any]:
        """
        Convert a quantum program from the source package to the target package.

        Args:
            program (qbraid.QPROGRAM): The quantum program to be converted.

        Returns:
            Union[qbraid.QPROGRAM, Any]: The converted quantum program,
                                         typically of a supported program type.

        Raises:
            ValueError: If the provided program's type does not match the source package type.
        """
        package = get_program_type(program)
        if package != self._source:
            raise ValueError(
                f"Expected program of type {SUPPORTED_QPROGRAMS[self._source]}, "
                f"but got program of type {SUPPORTED_QPROGRAMS[package]}."
            )
        return self._conversion_func(program)

    def __repr__(self) -> str:
        """
        Represent the Conversion instance as a string indicating
        source and target packages.

        Returns:
            str: String representation of the Conversion instance.
        """
        return f"('{self._source}', '{self._target}')"

    def __eq__(self, other: Any) -> bool:
        """
        Check if another instance is equal to this instance.

        Args:
            other (Any): Another instance to compare.

        Returns:
            bool: True if the instances are equal, False otherwise.
        """
        if not isinstance(other, Conversion):
            return False

        return (
            self._source == other._source
            and self._target == other._target
            and self._native == other._native
        )



# Copyright (C) 2023 qBraid
# Copyright (C) Unitary Fund
#
# This file is part of the qBraid-SDK.
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# This file includes code adapted from Mitiq (https://github.com/unitaryfund/mitiq)
# with modifications by qBraid. The original copyright notice is included above.
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# qbraid: skip-header

"""
Module containing function to test Cirq circuit equality

"""
from copy import deepcopy

import cirq

import qbraid.programs.cirq


def _rev_qubits(circuit: cirq.Circuit) -> cirq.Circuit:
    """Reverses the qubit order of a circuit."""
    program = qbraid.programs.cirq.CirqCircuit(circuit)
    program.reverse_qubit_order()
    return program.program


def _equal(
    circuit_one: cirq.Circuit,
    circuit_two: cirq.Circuit,
    require_qubit_equality: bool = False,
    require_measurement_equality: bool = False,
    allow_reversed_qubit_order: bool = False,
) -> bool:
    """Returns True if the circuits are equal, else False.

    Args:
        circuit_one: Input circuit to compare to circuit_two.
        circuit_two: Input circuit to compare to circuit_one.
        require_qubit_equality: Requires that the qubits be equal
            in the two circuits.
        require_measurement_equality: Requires that measurements are equal on
            the two circuits, meaning that measurement keys are equal.
        allow_reversed_qubit_order: Allows considering the circuits as equal
            even if their qubit order is reversed. Default is False.

    Note:
        If set(circuit_one.all_qubits()) = {LineQubit(0)},
        then set(circuit_two_all_qubits()) must be {LineQubit(0)},
        else the two are not equal.
        If True, the qubits of both circuits must have a well-defined ordering.
    """
    # Make a deepcopy only if it's necessary
    if not (require_qubit_equality and require_measurement_equality):
        circuit_one = deepcopy(circuit_one)
        circuit_two = deepcopy(circuit_two)

    if not require_qubit_equality:
        # Transform the qubits of circuit one to those of circuit two
        qubit_map = dict(
            zip(
                sorted(circuit_one.all_qubits()),
                sorted(circuit_two.all_qubits()),
            )
        )
        circuit_one = circuit_one.transform_qubits(lambda q: qubit_map[q])

    if not require_measurement_equality:
        for circ in (circuit_one, circuit_two):
            measurements = [
                (moment, op)
                for moment, op, _ in circ.findall_operations_with_gate_type(cirq.MeasurementGate)
            ]
            circ.batch_remove(measurements)

    if allow_reversed_qubit_order and circuit_one != circuit_two:
        reversed_circuit_one = _rev_qubits(circuit_one)
        return _equal(
            reversed_circuit_one,
            circuit_two,
            require_qubit_equality,
            require_measurement_equality,
            False,
        )

    return circuit_one == circuit_two



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for visualizing quantum programs, experimental results,
and other associated data.

.. currentmodule:: qbraid.visualization

Functions
----------

.. autosummary::
   :toctree: ../stubs/

   plot_histogram
   plot_distribution
   plot_conversion_graph
   circuit_drawer
   qasm3_drawer

Exceptions
-----------

.. autosummary::
   :toctree: ../stubs/

   VisualizationError
"""
from .draw_circuit import circuit_drawer
from .draw_qasm3 import qasm3_drawer
from .exceptions import VisualizationError
from .plot_conversions import plot_conversion_graph
from .plot_counts import plot_distribution, plot_histogram



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining OpenQasm2Program Class

"""
import re
from typing import TYPE_CHECKING, List

from qbraid.programs.abc_program import QuantumProgram

if TYPE_CHECKING:
    import numpy as np


class OpenQasm2Program(QuantumProgram):
    """Wrapper class for OpenQASM 2 strings."""

    def __init__(self, program: str):
        super().__init__(program)

    @property
    def program(self) -> str:
        return self._program

    @program.setter
    def program(self, value: str) -> None:
        if not isinstance(value, str):
            raise ValueError("Program must be an instance of str")
        self._program = value

    def _get_bits(self, bit_type: str) -> List[str]:
        """Return the number of qubits or classical bits in the circuit.

        Args:
            bit_type: either "q" or "c" for qubits or classical bits, respectively.

        """
        matches = re.findall(rf"{bit_type}reg (\w+)\[(\d+)\];", self.program)

        result = []
        for match in matches:
            var_name = match[0]
            n = int(match[1])
            result.extend([f"{var_name}[{i}]" for i in range(n)])

        return result

    @property
    def qubits(self) -> List[str]:
        """Use regex to extract all qreg definitions from the string"""
        return self._get_bits("q")

    @property
    def num_clbits(self) -> int:
        """Return the number of classical bits in the circuit."""
        return self._get_bits("c")

    @staticmethod
    def _get_max_count(counts_dict) -> int:
        return max(counts_dict.values()) if counts_dict else 0

    @property
    def depth(self) -> int:  # pylint: disable=too-many-statements
        """Calculates circuit depth of OpenQASM 2 string"""
        qasm_str = self.program
        lines = qasm_str.splitlines()

        # Keywords marking lines to ommit from depth calculation.
        not_counted = ("OPENQASM", "include", "qreg", "creg", "gate", "opaque", "//")
        gate_lines = [s for s in lines if s.strip() and not s.startswith(not_counted)]

        all_qubits = self.qubits
        depth_counts = {qubit: 0 for qubit in all_qubits}

        track_measured = {}

        for _, s in enumerate(gate_lines):
            if s.startswith("barrier"):
                max_depth = self._get_max_count(depth_counts)
                depth_counts = {key: max_depth for key in depth_counts.keys()}
                continue

            raw_matches = re.findall(r"(\w+)\[(\d+)\]", s)

            matches = []
            if len(raw_matches) == 0:
                try:
                    if s.startswith("measure"):
                        match = re.search(r"measure (\w+) -> .+", s)
                        if match:
                            op = match.group(1)
                        else:
                            continue
                    else:
                        op = s.split(" ")[-1].strip(";")
                    for qubit in all_qubits:
                        qubit_name = qubit.split("[")[0]
                        if op == qubit_name:
                            matches.append(qubit)
                # pylint: disable=broad-exception-caught
                except Exception:
                    continue

                if len(matches) == 0:
                    continue
            else:
                for match in raw_matches:
                    var_name = match[0]
                    n = int(match[1])
                    qubit = f"{var_name}[{n}]"
                    if qubit in all_qubits and qubit in depth_counts:
                        matches.append(qubit)

            if s.startswith("if"):
                match = re.search(r"if\((\w+)==\d+\)", s)
                if match:
                    creg = match.group(1)
                    if creg in track_measured:
                        meas_qubits, meas_depth = track_measured[creg]
                        max_measured = max(max(depth_counts[q] for q in meas_qubits), meas_depth)
                        track_measured[creg] = meas_qubits, max_measured + 1
                        qubit = matches[0]
                        depth_counts[qubit] = max(max_measured, depth_counts[qubit]) + 1
                        continue

            # Calculate max depth among the qubits in the current line.
            max_depth = 0
            for qubit in matches:
                max_depth = max(max_depth, depth_counts[qubit])

            # Update depths for all qubits in the current line.
            for qubit in matches:
                depth_counts[qubit] = max_depth + 1

            if s.startswith("measure"):
                match = re.search(r"measure .+ -> (\w+)", s)
                if match:
                    creg = match.group(1)
                    track_measured[creg] = matches, max_depth + 1

        return self._get_max_count(depth_counts)

    def _unitary(self) -> "np.ndarray":
        """Return the unitary of the QASM"""
        # pylint: disable=import-outside-toplevel
        from qbraid.transpiler.conversions.cirq.conversions_qasm import qasm2_to_cirq

        return qasm2_to_cirq(self.program).unitary()

    def remove_idle_qubits(self) -> None:
        """Checks whether the circuit uses contiguous qubits/indices,
        and if not, reduces dimension accordingly."""
        raise NotImplementedError

    def reverse_qubit_order(self) -> None:
        """Reverses the qubit ordering of a openqasm program."""
        raise NotImplementedError



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# isort: skip_file
# pylint: skip-file

"""
Mdule submiting and managing quantm tasks through AWS
and Amazon Braket supported devices.

.. currentmodule:: qbraid.providers.aws

Classes
---------

.. autosummary::
   :toctree: ../stubs/

   BraketDevice
   BraketProvider
   BraketQuantumTask
   BraketGateModelResult

Functions
-----------

.. autosummary::
   :toctree: ../stubs/

   get_quantum_task_cost

"""
from .result import BraketGateModelResult
from .device import BraketDevice
from .job import BraketQuantumTask
from .provider import BraketProvider
from .tracker import get_quantum_task_cost



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for plotting historgram of measurement counts against quantum states.

"""

from typing import Callable, Dict, List, Optional, Union

import matplotlib.pyplot as plt
from matplotlib import colormaps

# pylint: disable=too-many-arguments,unnecessary-lambda


def _counts_to_decimal(counts: dict) -> dict:
    """
    Converts a dictionary of counts to decimal form.

    Args:
        counts_dict (dict): A dictionary where keys are strings representing states and
                            values are integers representing counts.

    Returns:
        dict: A dictionary with the same keys as the input dictionary, but with values
              converted to their respective decimal proportions.

    Raises:
        ValueError: If the total count is zero.
        TypeError:  If the input dictionary contains non-integer values

    Example:
        >>> counts_to_decimal({"00": 10, "01": 15, "10": 20, "11": 5})
        {"00": 0.2, "01": 0.3, "10": 0.4, "11": 0.1}
    """

    try:
        total_count = sum(counts.values())
    except TypeError as err:
        raise TypeError("Counts values must be integers.") from err

    if total_count == 0:
        raise ValueError("Total count cannot be zero.")

    decimal_dict = {key: value / total_count for key, value in counts.items()}

    return decimal_dict


def _plot_data(
    counts: Union[List[Dict], dict],
    legend: Optional[Union[List[str], str]] = None,
    colors: Optional[Union[List[str], str]] = None,
    title: Optional[str] = None,
    x_label: Optional[str] = None,
    y_label: Optional[str] = None,
    show_plot: Optional[bool] = True,
    save_path: Optional[str] = None,
    transform_fn: Optional[Callable[[dict], dict]] = None,
    label_format_fn: Optional[Callable[[float], str]] = lambda x: str(x),
):
    if not isinstance(counts, list):
        counts = [counts]
        if isinstance(colors, str):
            colors = [colors]
        if isinstance(legend, str):
            legend = [legend]

    if transform_fn:
        counts = [transform_fn(counts_dict) for counts_dict in counts]

    all_states = sorted(set(state for counts_dict in counts for state in counts_dict.keys()))

    num_dicts = len(counts)
    bar_width = 0.8 / num_dicts

    x_positions = range(len(all_states))

    if colors is None:
        cmap = colormaps.get_cmap("tab10")
        colors = [cmap(i / 10) for i in range(num_dicts)]

    if len(colors) != len(counts):
        raise ValueError("Number of colors must match number of datasets")

    if isinstance(legend, list) and len(legend) != len(counts):
        raise ValueError("Number of legend labels must match number of datasets")

    for i, counts_dict in enumerate(counts):
        counts_iter = [counts_dict.get(state, 0) for state in all_states]

        default_label = f"Job {i}" if num_dicts > 1 else None
        label = legend[i] if legend and i < len(legend) else default_label

        bars = plt.bar(
            [x + (i * bar_width) for x in x_positions],
            counts_iter,
            width=bar_width,
            color=colors[i % num_dicts],
            label=label,
            align="center",
        )

        y_min, y_max = plt.gca().get_ylim()
        y_range = y_max - y_min
        offset_percentage = 0.02  # Adjust this value to get the desired offset
        y_offset = y_range * offset_percentage

        for hbar, count in zip(bars, counts_iter):
            plt.text(
                hbar.get_x() + hbar.get_width() / 2,
                hbar.get_height() + +y_offset,
                label_format_fn(count),
                ha="center",
                va="bottom",
                color="black",
                fontsize=8,
            )

    plt.xticks(x_positions, all_states, rotation=45)

    y_ticks = plt.gca().get_yticks()
    plt.yticks(y_ticks)
    plt.grid(axis="y", linestyle="--", linewidth=0.7, alpha=0.7)

    if y_label:
        plt.ylabel(y_label)

    if x_label:
        plt.xlabel(x_label)

    if title:
        plt.title(title)

    if legend or num_dicts > 1:
        plt.legend()

    if save_path:
        plt.savefig(save_path)

    if show_plot:
        plt.show()


def plot_distribution(
    counts: Union[List[Dict], Dict],
    legend: Optional[Union[List[str], str]] = None,
    colors: Optional[Union[List[str], str]] = None,
    title: Optional[str] = None,
    x_label: Optional[str] = None,
    y_label: Optional[str] = None,
    show_plot: Optional[bool] = True,
    save_path: Optional[str] = None,
):
    """
    Plots a histogram probability distribution of quantum states.

    Args:
        counts (Union[List[Dict], Dict]): Dictionary or a list of dictionaries containing the
                                          quantum states as keys and their respective counts as
                                          values.
        legend (Optional[Union[List[str], str]]): List of strings or a single string representing
                                                  the labels of the datasets. Defaults to None,
                                                  where it generates default labels.
        colors (Optional[Union[List[str], str]]): List of strings or a single string representing
                                                  the colors for each dataset. Defaults to None,
                                                  where it generates a color sequence.
        title (Optional[str]): String representing the title of the plot. Defaults to None.
        x_label (Optional[str]): String representing the label for the x-axis. Defaults to None.
        y_label (Optional[str]): String representing the label for the y-axis. Defaults to None.
        show_plot (Optional[bool]): Boolean representing whether to show the plot. Defaults to True.
        save_path (Optional[str]): String representing the path to save the plot. Defaults to None.

    Returns:
        None: This function does not return a value; it displays a plot.

    Raises:
        ValueError: Raises an error if input arguments do not match the expected types or formats.

    Example:
        .. code-block:: python

            counts_dict1 = {'00': 50, '01': 30, '10': 10, '11': 10}
            counts_dict2 = {'00': 20, '01': 40, '10': 30, '11': 10}

            plot_distribution(
                counts=[counts_dict1, counts_dict2],
                legend=['First Execution', 'Second Execution'],
                colors=['crimson', 'midnightblue'],
                title="Quantum State Probability Distribution",
                x_label="Quantum States",
                y_label="Probabilities"
            )
    """

    y_label = "Quasi-probabilities" if not y_label else y_label

    _plot_data(
        counts,
        legend,
        colors,
        title,
        x_label,
        y_label,
        show_plot,
        save_path,
        transform_fn=_counts_to_decimal,
        label_format_fn=lambda x: "{:.3f}".format(x),  # pylint: disable=consider-using-f-string
    )


def plot_histogram(
    counts: Union[List[Dict], Dict],
    legend: Optional[Union[List[str], str]] = None,
    colors: Optional[Union[List[str], str]] = None,
    title: Optional[str] = None,
    x_label: Optional[str] = None,
    y_label: Optional[str] = None,
    show_plot: Optional[bool] = True,
    save_path: Optional[str] = None,
):
    """
    Plots a histogram of measurement counts against quantum states.

    Args:
        counts (Union[List[Dict], Dict]): Dictionary or a list of dictionaries containing the
                                          quantum states as keys and their respective counts as
                                          values.
        legend (Optional[Union[List[str], str]]): List of strings or a single string representing
                                                  the labels of the datasets. Defaults to None,
                                                  where it generates default labels.
        colors (Optional[Union[List[str], str]]): List of strings or a single string representing
                                                  the colors for each dataset. Defaults to None,
                                                  where it generates a color sequence.
        title (Optional[str]): String representing the title of the plot. Defaults to None.
        x_label (Optional[str]): String representing the label for the x-axis. Defaults to None.
        y_label (Optional[str]): String representing the label for the y-axis. Defaults to None.
        show_plot (Optional[bool]): Boolean representing whether to show the plot. Defaults to True.
        save_path (Optional[str]): String representing the path to save the plot. Defaults to None.

    Returns:
        None: This function does not return a value; it displays a plot.

    Raises:
        ValueError: Raises an error if input arguments do not match the expected types or formats.

    Example:
        .. code-block:: python

            counts_dict1 = {'00': 50, '01': 30, '10': 10, '11': 10}
            counts_dict2 = {'00': 20, '01': 40, '10': 30, '11': 10}

            plot_histogram(
                counts=[counts_dict1, counts_dict2],
                legend=['First Execution', 'Second Execution'],
                colors=['crimson', 'midnightblue'],
                title="Quantum State Measurement Counts",
                x_label="Quantum States",
                y_label="Counts"
            )
    """

    y_label = "Counts" if not y_label else y_label

    _plot_data(
        counts,
        legend,
        colors,
        title,
        x_label,
        y_label,
        show_plot,
        save_path,
        label_format_fn=lambda x: str(x),
    )



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: skip-file

"""
Module for interfacing with the qBraid API.

.. currentmodule:: qbraid.api

Exceptions
------------

.. autosummary::
   :toctree: ../stubs/

   ApiError
   AuthError
   ConfigError
   RequestsApiError


Classes
--------

.. autosummary::
   :toctree: ../stubs/

   QbraidSession
   PostForcelistRetry

"""
from .exceptions import ApiError, AuthError, ConfigError, RequestsApiError
from .retry import PostForcelistRetry
from .session import QbraidSession



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for generating random Qiskit circuits

"""
from qiskit import QuantumCircuit
from qiskit.circuit.exceptions import CircuitError
from qiskit.circuit.random import random_circuit

from qbraid.exceptions import QbraidError


def _qiskit_random(num_qubits: int, depth: int, **kwargs) -> QuantumCircuit:
    """Generate random circuit qiskit circuit.

    Args:
        num_qubits (int): number of quantum wires
        depth (int): layers of operations (i.e. critical path length)

    Raises:
        QbraidError: When invalid qiskit random circuit options given

    Returns:
        Qiskit random circuit

    """
    if "measure" not in kwargs:
        kwargs["measure"] = False

    try:
        return random_circuit(num_qubits, depth, **kwargs)
    except CircuitError as err:
        raise QbraidError("Could not create Qiskit random circuit") from err



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for Amazon Braket cost tracker interface

"""
import os
from decimal import Decimal

import pytest
from braket.circuits import Circuit
from braket.tracking import Tracker

from qbraid.providers.aws import BraketProvider
from qbraid.providers.aws.tracker import get_quantum_task_cost

# Skip tests if AWS account auth/creds not configured
skip_remote_tests: bool = os.getenv("QBRAID_RUN_REMOTE_TESTS") is None
REASON = "QBRAID_RUN_REMOTE_TESTS not set (requires configuration of AWS storage)"


@pytest.mark.skipif(skip_remote_tests, reason=REASON)
def test_get_quantum_task_cost_simulator():
    """Test getting cost of quantum task run on an AWS simulator."""
    provider = BraketProvider()
    device = provider.get_device("arn:aws:braket:::device/quantum-simulator/amazon/sv1")
    circuit = Circuit().h(0).cnot(0, 1)

    with Tracker() as tracker:
        task = device.run(circuit, shots=2)
        task.result()

    expected = tracker.simulator_tasks_cost()
    calculated = get_quantum_task_cost(task.id, provider._get_aws_session())
    assert expected == calculated


@pytest.mark.skipif(skip_remote_tests, reason=REASON)
def test_get_quantum_task_cost_cancelled(braket_most_busy, braket_circuit):
    """Test getting cost of quantum task that was cancelled."""
    if braket_most_busy is None:
        pytest.skip("No AWS QPU devices available")

    provider = BraketProvider()

    # AwsSession region must match device region
    region_name = provider._get_region_name(braket_most_busy._id)
    aws_session = provider._get_aws_session(region_name)

    qbraid_job = braket_most_busy.run(braket_circuit, shots=10)
    qbraid_job.cancel()

    task_arn = qbraid_job.vendor_job_id

    # with pytest.raises(ValueError) as excinfo:
    #     get_quantum_task_cost(task_arn, aws_session)

    # assert "Current state is CANCELLING" in str(excinfo.value)

    qbraid_job.wait_for_final_state()
    assert get_quantum_task_cost(task_arn, aws_session) == Decimal(0)


@pytest.mark.skipif(skip_remote_tests, reason=REASON)
def test_get_quantum_task_cost_region_mismatch(braket_most_busy, braket_circuit):
    """Test getting cost of quantum task raises value error on region mismatch."""
    if braket_most_busy is None:
        pytest.skip("No AWS QPU devices available")

    braket_device = braket_most_busy._device
    task = braket_device.run(braket_circuit, shots=10)
    task.cancel()

    task_arn = task.id
    task_region = task_arn.split(":")[3]
    other_region = "eu-west-2" if task_region == "us-east-1" else "us-east-1"

    provider = BraketProvider()
    aws_session = provider._get_aws_session(other_region)

    with pytest.raises(ValueError) as excinfo:
        get_quantum_task_cost(task_arn, aws_session)

    assert (
        str(excinfo.value)
        == f"AwsSession region {other_region} does not match task region {task_region}"
    )



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining input / output types for a quantum backend:

  * QPROGRAM: Type alias defining all supported quantum circuit / program types

  * SUPPORTED_QPROGRAMS: Dict mapping all supported quantum software libraries / package
                         names to their respective program types

"""

from importlib import import_module
from types import ModuleType
from typing import List, Union

__NON_OPTIONAL_PROGRAMS: list = []

# As we are using here dynamic imports, we'll see the ide will visualize that none of the
# libraries have been import, as we'll import them dynamically in the execution.


def __dynamic_importer(opt_modules: List[str]) -> list:
    imported: list = __NON_OPTIONAL_PROGRAMS
    for m in opt_modules:
        try:
            data = m.split(".")
            for i, _ in enumerate(data):
                if data[:i]:
                    globals()[".".join(data[:i])] = import_module(".".join(data[:i]))
            # to be more secure, do not do module = globals()[m] = import_module(), as it could
            # create globals()[m] and later throw an error as there is no module named like that.
            # Is prefered to let module be import_module() and throw an exception if is needed
            module: ModuleType = import_module(m)
            globals()[m] = module
            imported.append(__get_class(module.__name__))
        except Exception:  # pylint: disable=broad-except
            pass
    return imported


# pylint: disable=undefined-variable,inconsistent-return-statements
def __get_class(module: str):
    if module == "cirq":
        return cirq.Circuit  # type: ignore
    if module == "qiskit":
        return qiskit.QuantumCircuit  # type: ignore
    if module == "braket.circuits":
        return braket.circuits.Circuit  # type: ignore
    if module == "pennylane":
        return pennylane.tape.QuantumTape  # type: ignore
    if module == "pyquil":
        return pyquil.Program  # type: ignore
    if module == "pytket":
        return pytket._tket.circuit.Circuit  # type: ignore
    if module == "openqasm3":
        return openqasm3.ast.Program  # type: ignore


# Supported quantum programs.
QASMType = str
_PROGRAMS = __dynamic_importer(
    ["cirq", "qiskit", "pennylane", "pyquil", "pytket", "braket.circuits", "openqasm3"]
)
QPROGRAM = Union[tuple(_PROGRAMS)]  # type: ignore

# pylint: disable-next=bad-str-strip-call
_PROGRAM_TYPES = [str(x).strip("<class").strip(">").strip(" ").strip("'") for x in _PROGRAMS]
QPROGRAM_TYPES = _PROGRAMS + [QASMType]

_PROGRAM_LIBS = [x.split(".")[0] for x in _PROGRAM_TYPES]
SUPPORTED_QPROGRAMS = dict(zip(_PROGRAM_LIBS, _PROGRAM_TYPES)) | {"qasm2": "str", "qasm3": "str"}
QPROGRAM_LIBS = list(SUPPORTED_QPROGRAMS.keys())



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for qbraid.programs.cirq.CirqCircuit

"""
import pytest
from cirq import CNOT, Circuit, GridQubit, H, LineQubit, Moment, NamedQubit, X, Y, Z

from qbraid.interface import circuits_allclose
from qbraid.programs.cirq import CirqCircuit


def test_contiguous_line_qubits():
    """Test convert to contiguous method for line qubits."""
    circuit = Circuit()
    circuit.append(X(LineQubit(0)))
    circuit.append(Y(LineQubit(2)))
    circuit.append(Z(LineQubit(4)))

    program = CirqCircuit(circuit)
    program.remove_idle_qubits()

    expected_qubits = [LineQubit(0), LineQubit(1), LineQubit(2)]
    assert set(program.qubits) == set(expected_qubits)


def test_remove_idle_grid_qubits_row():
    """Test convert to contiguous method for grid qubits."""
    circuit = Circuit()
    circuit.append(X(GridQubit(0, 0)))
    circuit.append(Y(GridQubit(0, 2)))
    circuit.append(Z(GridQubit(0, 4)))

    program = CirqCircuit(circuit)
    program.remove_idle_qubits()

    expected_qubits = [GridQubit(0, 0), GridQubit(0, 1), GridQubit(0, 2)]
    assert set(program.qubits) == set(expected_qubits)


@pytest.mark.skip(reason="Not implemented yet")
def test_remove_idle_grid_qubits_col():
    """Test convert to contiguous method for grid qubits."""
    circuit = Circuit()
    circuit.append(X(GridQubit(0, 0)))
    circuit.append(Y(GridQubit(2, 0)))
    circuit.append(Z(GridQubit(4, 0)))

    program = CirqCircuit(circuit)
    program.remove_idle_qubits()

    expected_qubits = [GridQubit(0, 0), GridQubit(1, 0), GridQubit(2, 0)]
    assert set(program.qubits) == set(expected_qubits)


@pytest.mark.skip(reason="Not implemented yet")
@pytest.mark.parametrize("index1,index2", [(1, 2), (2, 1)])
def test_remove_idle_qubits_raises(index1, index2):
    """Test removing idle qubits for grid qubits raises exception if
    row or col indicies are not all equal."""
    circuit = Circuit()
    circuit.append(X(GridQubit(0, 0)))
    circuit.append(Y(GridQubit(index1, index2)))

    program = CirqCircuit(circuit)

    with pytest.raises(ValueError):
        program.remove_idle_qubits()


@pytest.mark.skip(reason="Not implemented yet")
def test_convert_grid_to_line_qubits():
    """Test converting grid qubits to line qubits."""
    circuit = Circuit()
    circuit.append(X(GridQubit(0, 0)))
    circuit.append(Y(GridQubit(0, 1)))
    circuit.append(Z(GridQubit(0, 2)))

    program = CirqCircuit(circuit)
    program._convert_to_line_qubits()

    expected_qubits = [LineQubit(0), LineQubit(1), LineQubit(2)]
    assert set(program.qubits) == set(expected_qubits)


def test_mixed_qubit_types():
    """Test convert to contiguous method raises error for mixed qubit type."""
    circuit = Circuit()
    circuit.append(X(LineQubit(3)))
    circuit.append(Y(GridQubit(0, 4)))

    program = CirqCircuit(circuit)
    program.remove_idle_qubits()

    expected_qubits = [LineQubit(0), LineQubit(1)]
    assert set(program.qubits) == set(expected_qubits)


def test_convert_named_qubit_to_line_qubit():
    """Test converting a circuit of all NamedQubits to LineQubits"""
    circuit = Circuit(
        [
            Moment(H(NamedQubit("q"))),
        ]
    )

    program = CirqCircuit(circuit)
    program._convert_to_line_qubits()
    for qubit in program.qubits:
        assert isinstance(qubit, LineQubit)


def test_reverse_named_qubit_order():
    """Test reversing the order of qubits in a NamedQubit circuit"""
    circuit = Circuit(
        [
            Moment(H(NamedQubit("q"))),
            Moment(X(NamedQubit("y"))),
        ]
    )

    op_map = {}
    for op in circuit.all_operations():
        op_map[str(op.gate)] = op.qubits

    program = CirqCircuit(circuit)
    program.reverse_qubit_order()
    rev_circuit = program.program
    for op in rev_circuit.all_operations():
        assert op_map[str(op.gate)] != op.qubits


def test_remove_idle_named_qubits():
    """Test removing idle qubits from a NamedQubit circuit"""
    circuit = Circuit(
        [
            Moment(H(NamedQubit("q_4")), H(NamedQubit("q_8"))),
            Moment(CNOT(NamedQubit("q_6"), NamedQubit("q_8"))),
            Moment(X(NamedQubit("q_2"))),
        ]
    )
    qprogram = CirqCircuit(circuit)
    qprogram.remove_idle_qubits()
    new_circuit = qprogram.program
    assert set(new_circuit.all_qubits()) == {NamedQubit(str(i)) for i in range(qprogram.num_qubits)}
    assert circuits_allclose(circuit, new_circuit)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for qbraid.programs.braket.BraketCircuit

"""
from itertools import chain, combinations

import numpy as np
import pytest
from braket.circuits import Circuit, Instruction, gates

from qbraid.programs.braket import BraketCircuit


def get_subsets(nqubits):
    """Return list of all combinations up to number nqubits"""
    qubits = list(range(0, nqubits))

    def combos(x):
        return combinations(qubits, x)

    all_subsets = chain(*map(combos, range(0, len(qubits) + 1)))
    return list(all_subsets)[1:]


def calculate_expected(gates_set):
    """Calculated expected unitary"""
    if len(gates_set) == 1:
        return gates_set[0]
    if len(gates_set) == 2:
        return np.kron(gates_set[1], gates_set[0])
    return np.kron(calculate_expected(gates_set[2:]), np.kron(gates_set[1], gates_set[0]))


def generate_test_data(input_gate_set, contiguous=True):
    """Generate test data"""
    testdata = []
    gate_set = input_gate_set.copy()
    gate_set.append(gates.I)
    nqubits = len(input_gate_set)
    subsets = get_subsets(nqubits)
    for ss in subsets:
        bk_instrs = []
        np_gates = []
        for index in range(max(ss) + 1):
            idx = -1 if index not in ss else index
            BKGate = gate_set[idx]
            np_gate = BKGate().to_matrix()
            if idx != -1 or contiguous:
                bk_instrs.append((BKGate, index))
            np_gates.append(np_gate)
        u_expected = calculate_expected(np_gates)
        testdata.append((bk_instrs, u_expected))
    return testdata


def make_circuit(bk_instrs):
    """Constructs Braket circuit from list of instructions"""
    circuit = Circuit()
    for instr in bk_instrs:
        Gate, index = instr
        circuit.add_instruction(Instruction(Gate(), target=index))
    qprogram = BraketCircuit(circuit)
    qprogram.populate_idle_qubits()
    return qprogram.program


test_gate_set = [gates.X, gates.Y, gates.Z]
test_data_contiguous_qubits = generate_test_data(test_gate_set)
test_data_non_contiguous_qubits = generate_test_data(test_gate_set, contiguous=False)
test_data = test_data_contiguous_qubits + test_data_non_contiguous_qubits


@pytest.mark.parametrize("bk_instrs,u_expected", test_data)
def test_unitary_calc(bk_instrs, u_expected):
    """Test calculating unitary of circuits using both contiguous and
    non-contiguous qubit indexing."""
    circuit = make_circuit(bk_instrs)
    qbraid_circuit = BraketCircuit(circuit)
    u_test = qbraid_circuit.unitary_little_endian()
    assert np.allclose(u_expected, u_test)


@pytest.mark.parametrize("bk_instrs,u_expected", test_data)
def test_convert_be_to_le(bk_instrs, u_expected):
    """Test converting big-endian unitary to little-endian unitary."""
    circuit = make_circuit(bk_instrs)
    qbraid_circuit = BraketCircuit(circuit)
    u_test = qbraid_circuit.unitary_little_endian()
    assert np.allclose(u_expected, u_test)


def test_kronecker_product_factor_permutation():
    """Test calculating unitary permutation representing
    circuits with reversed qubits"""
    bk_circuit = Circuit().h(0).cnot(0, 1)
    circuit = BraketCircuit(bk_circuit)

    unitary = circuit.unitary()
    circuit.reverse_qubit_order()
    unitary_rev = circuit.unitary_rev_qubits()

    assert np.allclose(unitary, unitary_rev)


def test_remove_idle_qubits_braket_bell():
    """Test convert_to_contigious on bell circuit"""
    circuit = Circuit().h(0).cnot(0, 1)  # pylint: disable=no-member
    h_gate = np.sqrt(1 / 2) * np.array([[1, 1], [1, -1]])
    h_gate_kron = np.kron(np.eye(2), h_gate)
    cnot_gate = np.array([[1, 0, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0], [0, 1, 0, 0]])
    u_expected = np.einsum("ij,jk->ki", h_gate_kron, cnot_gate)
    qprogram = BraketCircuit(circuit)
    qprogram.remove_idle_qubits()
    u_little_endian = qprogram.unitary_little_endian()
    assert np.allclose(u_expected, u_little_endian)


def test_collapse_empty_braket_control_modifier():
    """Test that converting braket circuits to contiguous qubits works with control modifiers"""
    circuit = Circuit().y(target=0, control=1)
    qprogram = BraketCircuit(circuit)
    qprogram.remove_idle_qubits()
    contig_circuit = qprogram.program
    assert circuit.qubit_count == contig_circuit.qubit_count



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for converting qiskit circuits to/from OpenQASM.

"""
import pytest
from qiskit import QuantumCircuit

from qbraid.interface import circuits_allclose
from qbraid.transpiler.conversions.qiskit.conversions_qasm import (
    _add_stdgates_include,
    qasm3_to_qiskit,
    qiskit_to_qasm3,
)

qasm_stdgate_data = [
    (
        """
OPENQASM 3;
qubit[1] q;
h q[0];
rypi/4) q[0];
        """,
        """
OPENQASM 3;
include "stdgates.inc";
qubit[1] q;
h q[0];
rypi/4) q[0];
        """,
    ),
]


@pytest.mark.parametrize("qasm3_test_in, qasm3_expected_out", qasm_stdgate_data)
def test_add_stdgates_include(qasm3_test_in, qasm3_expected_out):
    """Test adding stdgates include to OpenQASM 3.0 string"""
    assert _add_stdgates_include(qasm3_test_in) == qasm3_expected_out


def test_qiskit_to_from_qasm3():
    """Test converting qiskit circuit to/from OpenQASM 3.0 string"""
    circuit_in = QuantumCircuit(2)
    circuit_in.h(0)
    circuit_in.cx(0, 1)

    qasm3_str = qiskit_to_qasm3(circuit_in)
    circuit_out = qasm3_to_qiskit(qasm3_str)
    assert circuits_allclose(circuit_in, circuit_out, strict_gphase=True)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.
"""
Module for transpiling quantum programs between different quantum programming languages

"""
import logging
import warnings
from copy import deepcopy
from typing import TYPE_CHECKING, Callable, List, Optional

from qbraid._qprogram import QPROGRAM_LIBS
from qbraid.inspector import get_program_type
from qbraid.transpiler.exceptions import (
    CircuitConversionError,
    ConversionPathNotFoundError,
    NodeNotFoundError,
)
from qbraid.transpiler.graph import ConversionGraph

if TYPE_CHECKING:
    import cirq

    import qbraid


def _warn_if_unsupported(program_type, program_direction):
    if program_type not in QPROGRAM_LIBS:
        warnings.warn(
            f"Converting {program_direction} unsupported program type '{program_type}'.",
            UserWarning,
        )


def _flatten_cirq(circuit: "cirq.Circuit") -> "cirq.Circuit":
    """
    Flatten a Cirq circuit.

    Args:
        circuit (cirq.Circuit): The Cirq circuit to flatten.

    Returns:
        cirq.Circuit: The flattened Cirq circuit.
    """
    # TODO: potentially replace with native cirq.decompose
    # https://quantumai.google/reference/python/cirq/decompose

    # pylint: disable=import-outside-toplevel
    from cirq.contrib.qasm_import import circuit_from_qasm

    return circuit_from_qasm(circuit.to_qasm())


def _get_path_from_bound_methods(bound_methods: List[Callable]) -> str:
    """
    Constructs a path string from a list of bound methods of Conversion instances.

    This function takes a list of bound methods (specifically 'convert' methods bound to
    Conversion instances) and constructs a path string representing the sequence of
    conversions. Each conversion is defined by the 'source' and 'target' properties of the
    Conversion instance to which each method is bound.

    Args:
        bound_methods: A list of bound methods from Conversion instances.

    Returns:
        A string representing the path of conversions, formatted as
        'source1 -> source2 -> ... -> targetN'.

    Raises:
        AttributeError: If the bound methods do not have the expected 'source'
                        and 'target' attributes.
        IndexError: If the list of bound methods is empty.
    """
    if not bound_methods:
        raise IndexError("The list of bound methods is empty.")

    path = []
    for method in bound_methods:
        instance = method.__self__  # Get the instance to which the method is bound
        if not hasattr(instance, "source") or not hasattr(instance, "target"):
            raise AttributeError("Bound method instance lacks 'source' or 'target' attributes.")
        path.append(instance.source)  # Add the source node of the instance

    # Add the target of the last method's instance to complete the path
    path.append(bound_methods[-1].__self__.target)

    return " -> ".join(path)


def convert_to_package(
    program: "qbraid.QPROGRAM",
    target: str,
    conversion_graph: Optional[ConversionGraph] = None,
    max_path_attempts: int = 3,
    max_path_depth: Optional[int] = None,
) -> "qbraid.QPROGRAM":
    """
    Transpile a quantum program to a target language using a conversion graph.
    This function attempts to find a conversion path from the program's current
    format to the target format. It can limit the search to a certain number of
    attempts and path depths.

    Args:
        program (qbraid.QPROGRAM): The quantum program to transpile.
        target (str): The target language to transpile to.
        conversion_graph (Optional[ConversionGraph]): The graph representing available conversions.
            If None, a default graph is used. Defaults to None.
        max_path_attempts (int): The maximum number of conversion paths to attempt before raising an
            exception. This is useful to avoid excessive computations when multiple paths are
            available. Defaults to 3.
        max_path_depth (Optional[int]): The maximum depth of conversions within a given path to
            allow. For example, a path with a depth of 2 would be ['cirq' -> 'qasm2' -> 'qiskit'],
            whereas a depth  of 1 would be a direct conversion ['cirq' -> 'braket']. Defaults
            to None, i.e. no limit set on the path depth.

    Returns:
        qbraid.QPROGRAM: The transpiled quantum program.

    Raises:
        NodeNotFoundError: If the target or source package is not in the ConversionGraph.
        ConversionPathNotFoundError: If no path is available to conversion between the
            source and target packages.
        CircuitConversionError: If the conversion fails through all attempted paths.
    """
    graph = conversion_graph or ConversionGraph()
    graph_type = "Default" if conversion_graph is None else "Provided"

    if not graph.has_node(target):
        raise NodeNotFoundError(graph_type, target, graph.nodes)

    source = get_program_type(program, require_supported=conversion_graph is None)

    if not graph.has_node(source):
        raise NodeNotFoundError(graph_type, target, graph.nodes)

    if not graph.has_path(source, target):
        raise ConversionPathNotFoundError(source, target)

    if source == target:
        return program

    _warn_if_unsupported(source, "from")
    _warn_if_unsupported(target, "to")

    paths = graph.find_top_shortest_conversion_paths(source, target, top_n=max_path_attempts)

    if max_path_depth is not None:
        paths = [path for path in paths if len(path) <= max_path_depth]
        if len(paths) == 0:
            raise ConversionPathNotFoundError(source, target, max_path_depth)

    for path in paths:
        temp_program = deepcopy(program)
        try:
            for convert_func in path:
                try:
                    temp_program = convert_func(temp_program)
                except Exception:  # pylint: disable=broad-exception-caught
                    if get_program_type(temp_program) == "cirq":
                        temp_program = _flatten_cirq(temp_program)
                        temp_program = convert_func(temp_program)  # Retry conversion
                    else:
                        raise
            logging.info(
                "\nSuccessfully transpiled using conversions: %s",
                _get_path_from_bound_methods(path),
            )
            return temp_program
        except Exception:  # pylint: disable=broad-exception-caught
            logging.info(
                "\nFailed to transpile using conversions: %s",
                _get_path_from_bound_methods(path),
            )
            continue

    raise CircuitConversionError(f"Failed to convert program from '{source}' to '{target}'.")



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining supported quantum device types:

  * QDEVICE: Type alias defining all supported quantum device / backend types

"""

from importlib import import_module
from types import ModuleType
from typing import List, Union

# As we are using here dynamic imports, we'll see the ide will visualize that none of the
# libraries have been import, as we'll import them dynamically in the execution.


def __dynamic_importer(opt_modules: List[str]) -> list:
    imported: list = []
    for m in opt_modules:
        try:
            data = m.split(".")
            for i, _ in enumerate(data):
                if data[:i]:
                    globals()[".".join(data[:i])] = import_module(".".join(data[:i]))
            # to be more secure, do not do module = globals()[m] = import_module(), as it could
            # create globals()[m] and later throw an error as there is no module named like that.
            # Is prefered to let module be import_module() and throw an exception if is needed
            module: ModuleType = import_module(m)
            globals()[m] = module
            imported.append(__get_class(module.__name__))
        except Exception:  # pylint: disable=broad-except
            pass
    return imported


# pylint: disable=undefined-variable,inconsistent-return-statements
def __get_class(module: str):
    if module == "qiskit_ibm_provider":
        return qiskit_ibm_provider.IBMBackend  # type: ignore
    if module == "qiskit_ibm_runtime":
        return qiskit_ibm_runtime.IBMBackend  # type: ignore
    if module == "braket.aws":
        return braket.aws.AwsDevice  # type: ignore


# Supported quantum devices.
_DEVICES = __dynamic_importer(["qiskit_ibm_provider", "qiskit_ibm_runtime", "braket.aws"])

# pylint: disable-next=line-too-long
QDEVICE = None if not _DEVICES else _DEVICES[0] if len(_DEVICES) == 1 else Union[tuple(_DEVICES)]  # type: ignore

# pylint: disable-next=bad-str-strip-call
QDEVICE_TYPES = [str(x).strip("<class").strip(">").strip(" ").strip("'") for x in _DEVICES]
_DEVICE_LIBS = [x.split(".")[0] for x in QDEVICE_TYPES]
SUPPORTED_QDEVICES = dict(zip(_DEVICE_LIBS, QDEVICE_TYPES))
QDEVICE_LIBS = list(SUPPORTED_QDEVICES.keys())



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module providing unified conversions interface between supported
quantum program types.

.. currentmodule:: qbraid.transpiler

Classes
--------

.. autosummary::
   :toctree: ../stubs/

   ConversionGraph
   Conversion

Functions
-----------

.. autosummary::
   :toctree: ../stubs/

   convert_to_package

Exceptions
-----------

.. autosummary::
   :toctree: ../stubs/

   CircuitConversionError
   NodeNotFoundError
   ConversionPathNotFoundError

"""
from .converter import convert_to_package
from .edge import Conversion
from .exceptions import CircuitConversionError, ConversionPathNotFoundError, NodeNotFoundError
from .graph import ConversionGraph



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining Qiskit OpenQASM conversions

"""
import qiskit
from qiskit.qasm3 import QASM3ImporterError, dumps, loads

gate_defs = {
    "iswap": """
gate iswap _gate_q_0, _gate_q_1 {
  s _gate_q_0;
  s _gate_q_1;
  h _gate_q_0;
  cx _gate_q_0, _gate_q_1;
  cx _gate_q_1, _gate_q_0;
  h _gate_q_1;
}""",
    "sxdg": """
gate sxdg _gate_q_0 {
  s _gate_q_0;
  h _gate_q_0;
  s _gate_q_0;
}""",
}


def _insert_gate_defs(qasm3_str: str) -> str:
    """Add gate definitions to a QASM 3.0 string.

    Args:
        qasm3_str (str): QASM 3.0 string.
    Returns:
        str: QASM 3.0 string with gate definitions.
    """
    lines = qasm3_str.splitlines()

    insert_index = 0
    for i, line in enumerate(lines):
        if "include" in line or "OPENQASM" in line:
            insert_index = i + 1

    for gate, defn in gate_defs.items():
        if gate in qasm3_str:
            lines.insert(insert_index, defn.strip())

    return "\n".join(lines)


def _add_stdgates_include(qasm_str: str) -> str:
    """Add 'include "stdgates.inc";' to the QASM string if it is missing."""
    if 'include "stdgates.inc";' in qasm_str:
        return qasm_str

    lines = qasm_str.splitlines()

    for i, line in enumerate(lines):
        if "OPENQASM" in line:
            lines.insert(i + 1, 'include "stdgates.inc";')
            break

    return "\n".join(lines)


def qasm3_to_qiskit(qasm3: str) -> qiskit.QuantumCircuit:
    """Convert QASM 3.0 string to a Qiskit QuantumCircuit representation.

    Args:
        qasm3 (str): A string in QASM 3.0 format.

    Returns:
        qiskit.QuantumCircuit: A QuantumCircuit object representing the input QASM 3.0 string.
    """
    # extra trailing space is intentional
    replacements = {
        "cnot ": "cx ",
        "si ": "sdg ",
        "ti ": "tdg ",
        "v ": "sx ",
        "vi ": "sxdg ",
        "phaseshift": "p",
        "cphaseshift": "cp",
    }

    def replace_commands(qasm_str, replacements):
        for old, new in replacements.items():
            qasm_str = qasm_str.replace(old, new)
        return qasm_str

    try:
        return loads(qasm3)
    except QASM3ImporterError:
        pass

    qasm3 = replace_commands(qasm3, replacements)
    qasm3 = _add_stdgates_include(qasm3)
    qasm3 = _insert_gate_defs(qasm3)

    return loads(qasm3)


def qiskit_to_qasm3(circuit: qiskit.QuantumCircuit) -> str:
    """Convert qiskit QuantumCircuit to QASM 3.0 string"""
    return dumps(circuit)


def qasm2_to_qiskit(qasm: str) -> qiskit.QuantumCircuit:
    """Returns a Qiskit circuit equivalent to the input OpenQASM 2 string.

    Args:
        qasm: OpenQASM 2 string to convert to a Qiskit circuit.

    Returns:
        Qiskit.QuantumCircuit object equivalent to the input OpenQASM 2 string.
    """
    return qiskit.QuantumCircuit.from_qasm_str(qasm)


def qiskit_to_qasm2(circuit: qiskit.QuantumCircuit) -> str:
    """Returns OpenQASM 2 string equivalent to the input Qiskit circuit.

    Args:
        circuit: Qiskit circuit to convert to OpenQASM 2 string.

    Returns:
        str: OpenQASM 2 representation of the input Qiskit circuit.
    """
    return circuit.qasm()



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for Amazon Braket circuit compilation

"""
import braket
import pytest

from qbraid.compiler.braket import braket_ionq_compile

from ..fixtures.braket.gates import get_braket_gates

braket_gates = get_braket_gates()


@pytest.mark.parametrize("gate_name", braket_gates)
def test_braket_ionq_compilation(gate_name):
    """Test converting Amazon Braket circuit to use only IonQ supprted gates"""
    gate = braket_gates[gate_name]
    if gate.qubit_count == 1:
        source_circuit = braket.circuits.Circuit([braket.circuits.Instruction(gate, 0)])
    else:
        source_circuit = braket.circuits.Circuit(
            [braket.circuits.Instruction(gate, range(gate.qubit_count))]
        )

        braket_circuit = braket_ionq_compile(source_circuit)
        assert isinstance(braket_circuit, braket.circuits.Circuit)
